import json
import shutil
from datetime import datetime

import attr
from PyQt5.QtWidgets import QTreeWidgetItem, QFileDialog
from botocore.exceptions import EndpointConnectionError
from epyqlib.device import DeviceInterface
from epyqlib.tabs.files.activity_log import ActivityLog, Event
from epyqlib.tabs.files.activity_syncer import ActivitySyncer
from epyqlib.tabs.files.association_cache import AssociationCache
from epyqlib.tabs.files.aws_login_manager import AwsLoginManager
from epyqlib.tabs.files.bucket_manager import BucketManager
from epyqlib.tabs.files.files_manager import FilesManager
from epyqlib.tabs.files.filesview import Cols, get_values
from epyqlib.tabs.files.log_manager import LogManager, PendingLog
from epyqlib.tabs.files.sync_config import SyncConfig, Vars
from epyqlib.utils.twisted import errbackhook
import twisted.internet
from twisted.internet.error import DNSLookupError
from twisted.internet.task import deferLater
from typing import Dict

from .graphql import API, InverterNotFoundException


@attr.s(slots=True)
class AssociationMapping:
    association = attr.ib()
    row: QTreeWidgetItem = attr.ib()


class FilesController:
    _tag = "[Files Controller]"
    from epyqlib.tabs.files.filesview import FilesView

    def __init__(self, view: FilesView):
        self.view = view
        self.old_notes: str = ""
        self._last_sync: datetime = None

        self.sync_config = SyncConfig.get_instance()
        self.api = API(self.sync_config.get_env())

        self.activity_log = ActivityLog()
        self.activity_syncer = ActivitySyncer(self.activity_log, self.api)
        self.association_cache = AssociationCache.init(self.sync_config.cache_dir)
        self.aws_login_manager = AwsLoginManager.get_instance()
        self.bucket_manager = BucketManager()
        self.cache_manager = FilesManager(self.sync_config.cache_dir)
        self.log_manager = LogManager.init(self.sync_config.cache_dir)

        self._is_offline = self.sync_config.get(Vars.offline_mode) or False
        self._is_connected = False
        self._serial_number = None
        self._inverter_id = None
        self._build_hash = None

        self._device_interface = None

        self._inverter_id_lookup: Dict[str, str] = {}  # serial -> inverter_id

        self.associations: [str, AssociationMapping] = {}
        self._log_rows: Dict[str, QTreeWidgetItem] = {}

    def setup(self):
        self.aws_login_manager.register_listener(self.login_status_changed)

        if self.sync_config.get(Vars.offline_mode):
            print(
                f"{self._tag} 'offline_mode' flag set to true. Simulating offline mode."
            )
            self.set_offline(True)

        self.view.bind()
        self.view.populate_tree()
        self.view.initialize_ui()

        self.view.show_logged_out_warning()
        self._ensure_current_token()

        self.activity_log.register_listener(
            lambda event: self.view.add_log_line(LogRenderer.render_event(event))
        )
        self.activity_log.register_listener(self.activity_syncer.listener)

        self.log_manager.add_listener(self._on_new_pending_log)

        self._show_pending_logs()

    def _ensure_current_token(self):

        if not self.aws_login_manager.is_logged_in():
            return

        if not self.aws_login_manager.is_session_valid():
            try:
                self.aws_login_manager.refresh()

                # Make sure that using the refresh token worked
                if not self.aws_login_manager.is_logged_in():
                    return

            except (EndpointConnectionError, DNSLookupError):
                print(
                    f"{self._tag} Unable to login to AWS. Setting offline mode to true."
                )
                self.set_offline(True)
                return

        self.set_offline(False)
        self.api.set_id_token(self.aws_login_manager.get_id_token())
        self.periodically_refresh_token()
        self.view.show_logged_in_status(True, self.aws_login_manager.get_username())

    def device_interface_set(self, device_interface: DeviceInterface):
        self._device_interface = device_interface

    async def _read_info_from_inverter(self):
        """
        Read serial number and build hash from the inverter, and query the API for the inverterId for
        the given serial number
        """
        # Hasn't been set yet
        if self._device_interface is None:
            return

        # Don't do this if we already have this information
        if self._inverter_id is not None:
            return

        bus_status: DeviceInterface.TransmitStatus = (
            self._device_interface.get_connected_status()
        )
        # Appears you can never be "connected" to the offline inverter?
        if not (bus_status.connected and bus_status.transmitting):
            return

        # Only read serial number if it wasn't set through the UI (Will this ever happen if transmit is true?)
        if self._serial_number is None:
            self._serial_number = await self._device_interface.get_serial_number()
            self._build_hash = await self._device_interface.get_build_hash()

            self.log_manager.build_id = self._build_hash

        await self._get_id_for_serial_number(self._serial_number)

    ## Sync Info Methods
    def _set_sync_time(self) -> None:
        self._last_sync = datetime.now()

    ## Data fetching
    async def get_inverter_associations(self, serial_number: str):
        """
        :raises InverterNotFoundException
        """

        if self._is_offline:
            associations = self.association_cache.get_associations(serial_number) or []
        else:
            associations = await self.api.get_associations(serial_number)
            self.association_cache.put_associations(serial_number, associations)

        for association in associations:
            if association["file"] is None:
                print(
                    f"[Files Controller] WARNING: Association {association['id']} returned with null file associated"
                )
                continue

            key = association["id"] + association["file"]["id"]
            if key in self.associations:
                self.associations[
                    key
                ].association = (
                    association  # Update the association in case it's changed
                )
                row: QTreeWidgetItem = self.associations[key].row
                self.view.ensure_correct_parent_for_row(
                    row, association["file"]["type"]
                )
            else:
                row = self.view.attach_row_to_parent(
                    association["file"]["type"], association["file"]["filename"]
                )
                self.view.show_question_icon(row, Cols.local)
                self.view.show_check_icon(row, Cols.web)
                self.associations[
                    association["id"] + association["file"]["id"]
                ] = AssociationMapping(association, row)

            # Render either the new or updated association
            self.view.render_association_to_row(association, row)

        # to_remove = [key for key, value in self.associations.items() if (value.association not in associations)]
        to_remove = []
        for key, mapping in self.associations.items():
            if mapping.association not in associations:
                to_remove.append(key)
        for key in to_remove:
            map: AssociationMapping = self.associations[key]
            map.row.parent().removeChild(map.row)
            del self.associations[key]

        self.view.sort_grid_items()
        self._set_sync_time()
        self.view.show_sync_time(self._last_sync)

    async def _sync_files(self):
        missing_hashes = set()

        mapping: AssociationMapping
        for key, mapping in self.associations.items():

            hash = mapping.association["file"]["hash"]
            if self.cache_manager.has_hash(hash):
                self.view.show_check_icon(mapping.row, Cols.local)
            else:
                # Don't proactively cache raw log files
                if mapping.association["file"]["type"].lower() == "log":
                    self.view.show_cross_status_icon(mapping.row, Cols.local)
                    continue
                missing_hashes.add(hash)

        if len(missing_hashes) == 0:
            print(f"{self._tag} All files already hashed locally.")
            return

        if self._is_offline:
            print(
                f"{self._tag} Not syncing missing files because we are currently offline."
            )
            return

        # TODO: Figure out how to download multiple files at a time. Just trying to wrap in asyncio task fails.
        for hash in missing_hashes:
            await self.sync_file(hash)

    async def sync_file(self, hash):
        await self.download_file(hash)

        mapping: AssociationMapping
        for mapping in self._get_mapping_for_hash(hash):
            self.view.show_check_icon(mapping.row, Cols.local)

    def _get_key_for_file_id(self, file_id: str):
        value: AssociationMapping
        return next(
            key
            for key, value in self.associations.items()
            if value.association["file"]["id"] == file_id
        )

    def _get_mapping_for_hash(self, hash: str) -> [AssociationMapping]:
        map: AssociationMapping
        return [
            map
            for map in self.associations.values()
            if map.association["file"]["hash"] == hash
        ]

    def get_hash_for_row(self, row: QTreeWidgetItem):
        return self._get_mapping_for_row(row).association["file"]["hash"]

    def _get_mapping_for_row(self, row: QTreeWidgetItem) -> AssociationMapping:
        try:
            return next(map for map in self.associations.values() if map.row == row)
        except StopIteration:
            return None

    def is_file_cached_locally(self, item: QTreeWidgetItem):
        hash = self._get_mapping_for_row(item).association["file"]["hash"]
        return self.cache_manager.has_hash(hash)

    async def download_file_for_row(self, row: QTreeWidgetItem):
        await self.download_file(
            self._get_mapping_for_row(row).association["file"]["hash"]
        )

    async def download_file(self, hash: str):
        print(f"[Files Controller] Downloading missing file hash {hash}")
        filename = self.cache_manager.get_file_path(hash)
        await self.bucket_manager.download_file(hash, filename)
        return hash

    async def download_log(self, hash: str):
        print(f"[Files Controller] Downloading missing log hash {hash}")
        filename = self.cache_manager.get_file_path(hash)
        await self.bucket_manager.download_log(hash, filename)
        return hash

    ## Lifecycle events
    async def tab_selected(self):
        self.cache_manager.verify_cache()

        self.view.serial_number.setText(self._serial_number)
        self.view.disable_serial_number_input(
            not self._device_interface.on_offline_bus()
        )

        self._show_pending_logs()

        if self.aws_login_manager.is_logged_in():
            self._ensure_current_token()

            if not self.aws_login_manager.is_logged_in():
                self.view.show_logged_out_warning()
                return

            if self.sync_config.get(Vars.auto_sync):
                await self.sync_now()

    async def on_bus_status_changed(self):
        # Note: this gets called at least once for the offline inverter before on_offline_bus is
        # set to true. So do not trust that exclusively. (BSB 18Apr19)
        if self._device_interface is not None:
            bus_status: DeviceInterface.TransmitStatus = (
                self._device_interface.get_connected_status()
            )
            if bus_status.connected and bus_status.transmitting:
                await self.sync_now()

    ## UI Events
    async def login_clicked(self):
        try:
            self.aws_login_manager.show_login_window(self.view.files_grid)
        except EndpointConnectionError:
            print(f"{self._tag} Unable to login to AWS. Setting offline mode to true.")
            self.set_offline(True)

    async def logout_clicked(self):
        self.aws_login_manager.log_user_out()
        print(f"{self._tag} Logging out from AWS. Setting offline mode to true.")
        self.set_offline(True)

    async def serial_number_from_parameters_clicked(self):
        self._serial_number = self._device_interface.get_serial_number_from_parameters()
        self.view.serial_number.setText(self._serial_number)
        await self._get_id_for_serial_number(self._serial_number)

    def open_file(self, row: QTreeWidgetItem):
        file_hash = self._get_mapping_for_row(row).association["file"]["hash"]
        file_path = self.cache_manager.get_file_path(file_hash)
        import subprocess, os, sys

        if sys.platform.startswith("darwin"):
            subprocess.call(("open", file_path))
        elif os.name == "nt":  # For Windows
            # noinspection PyUnresolvedReferences
            os.startfile(file_path)
        elif os.name == "posix":  # For Linux, Mac, etc.
            subprocess.call(("xdg-open", file_path))

    async def save_file_as_clicked(self, item: QTreeWidgetItem = None):
        if item is None:
            item = self.view.files_grid.currentItem()

        map = self._get_mapping_for_row(item)
        hash = map.association["file"]["hash"]
        if hash not in self.cache_manager.hashes():
            await self.sync_file(hash)

        (destination, filter) = QFileDialog.getSaveFileName(
            parent=self.view.files_grid,
            caption="Pick location to save file",
            directory=map.association["file"]["filename"],
        )

        if destination != "":
            shutil.copy2(self.cache_manager.get_file_path(hash), destination)

    async def sync_now(self):
        self.view.show_inverter_error(None)

        if self._serial_number is None:
            if self._device_interface.on_offline_bus():
                return

            if not self._device_interface.get_connected_status().connected:
                self.view.show_inverter_error(
                    "Bus is disconnected. Unable to get Inverter Serial Number"
                )
                return

            # If InverterId is not set and we're connected, get inverter serial #
            await self._read_info_from_inverter()

            if self._serial_number is None:
                # If we still can't get it, bail out
                return

            self.view.serial_number.setText(self._serial_number)

        self._ensure_current_token()

        ## Verify that we actually are offline
        if self._is_offline:
            try:
                await self.api.test_connection()
                self.set_offline(False)
            except Exception:
                pass

        try:
            await self.get_inverter_associations(self._serial_number)
        except InverterNotFoundException:
            self.view.show_inverter_error("Error: Inverter ID not found.")
            return
        except DNSLookupError:
            self.set_offline(True)
            return

        try:
            await self._sync_pending_logs()
        except EndpointConnectionError:
            self.set_offline(True)

        await self._sync_files()
        if not self._is_offline:
            await self.api.unsubscribe()
            await self.api.subscribe(
                self.aws_login_manager.get_user_customer(), self.subscription_fired
            )

    async def sync_all(self):
        self.view.add_log_line("Starting to sync all associations for organization.")

        # Pull associations
        all_associations = await self.api.get_associations_for_customer()

        # Write fresh associations to cache
        for serial, association_list in all_associations.items():
            self.association_cache.put_associations(serial, association_list)

        # Fetch missing files
        for hash in self.association_cache.get_all_known_file_hashes():
            if not self.cache_manager.has_hash(hash):
                try:
                    await self.download_file(hash)
                except Exception:
                    self.view.add_log_error_line(
                        f"Error caching file {hash}. See epyq.log for details."
                    )

        self.view.add_log_line("Completed syncing all associations for organization.")

    async def subscription_fired(self, action, payload):
        if action == "associationCreated":
            await self.show_new_association(payload["id"], payload["file"]["id"])
            print(
                f"{self._tag} Received association action {action}: {json.dumps(payload)}"
            )
            return
        elif action == "associationDeleted":
            key = self._get_key_for_file_id(payload["file"]["id"])
            self.view.remove_row(self.associations[key].row)
            del self.associations[key]
            return

        if "id" not in payload:
            raise Exception(
                f"ERROR: Unable to handle file subscription message."
                f"Payload doesn't contain file id.\n{json.dumps(payload, indent=2)}"
            )

        key = self._get_key_for_file_id(payload["id"])

        if action == "fileUpdated":
            map: AssociationMapping = self.associations[key]
            map.association["file"].update(payload)
            self.view.render_association_to_row(map.association, map.row)

        if action == "fileDeleted":
            self.view.remove_row(self.associations[key].row)
            del self.associations[key]

        value: AssociationMapping
        new_associations = [
            value.association for key, value in self.associations.items()
        ]

        self.association_cache.put_associations(self._serial_number, new_associations)

    async def show_new_association(self, association_id: str, file_id: str):
        association = await self.api.get_association(association_id, file_id)

        row = self.view.attach_row_to_parent(
            association["file"]["type"], association["file"]["filename"]
        )
        self.view.show_question_icon(row, Cols.local)
        self.view.show_check_icon(row, Cols.web)
        self.associations[
            association["id"] + association["file"]["id"]
        ] = AssociationMapping(association, row)
        self.view.render_association_to_row(association, row)
        await self.sync_file(association["file"]["hash"])

    def file_item_clicked(self, item: QTreeWidgetItem, column: int):
        if item in get_values(self.view.section_headers):
            self.view.show_file_details(None)
            return

        file_mapping: AssociationMapping = self._get_mapping_for_row(item)

        association = file_mapping.association

        readonly_description = (
            association["file"]["owner"] == "epc"
            and not self.aws_login_manager._cognito_helper.is_user_epc()
        )
        readonly_description = readonly_description or self._is_offline

        if file_mapping is not None:
            self.view.show_file_details(file_mapping.association, readonly_description)

            if self._is_offline:
                self.view.description.setReadOnly(True)
                self.view.notes.setReadOnly(True)

        if item in self._log_rows.values():
            hash = next(key for key, value in self._log_rows.items() if value == item)

    async def send_dummy_param_event(self, row: QTreeWidgetItem):
        map = self._get_mapping_for_row(row)

        inverter_id = await self._get_id_for_serial_number(self._serial_number)

        file = map.association["file"]
        event = Event.new_load_param_file(
            inverter_id, "PlaceholderUser", file["id"], file["hash"], file["filename"]
        )
        await self.activity_log.add(event)

    ## Application events
    async def login_status_changed(self, logged_in: bool):
        self.view.btn_sync_now.setDisabled(not logged_in)

        if logged_in:
            self.api.set_id_token(self.aws_login_manager.get_id_token())
            self.view.show_logged_in_status(True, self.aws_login_manager.get_username())
            self.periodically_refresh_token()
            await self.sync_now()
        else:
            self.view.show_logged_out_warning()
            self.association_cache.clear()
            await self.api.unsubscribe()

    def periodically_refresh_token(self):
        def _refresh():
            print(f"{self._tag} Refreshing access token.")
            self.aws_login_manager.refresh(force=True)
            self.api.set_id_token(self.aws_login_manager.get_id_token())
            deferLater(
                twisted.internet.reactor,
                self.aws_login_manager._cognito_helper._expires_in,
                _refresh,
            )

        _refresh()

    ## Notes
    def set_original_notes(self, description: str, notes: str):
        description = description or ""
        notes = notes or ""
        self.old_notes = notes + description

    def notes_modified(self, new_desc: str, new_notes: str):
        new_notes = new_notes + new_desc

        return (len(self.old_notes) != len(new_notes)) or self.old_notes != new_notes

    async def save_notes(self, file_id: str, description: str, notes: str):
        print(f"{self._tag} Saving updated notes for file {file_id}")
        await self.api.set_file_notes(file_id, description, notes)
        self.set_original_notes(description, notes)

    ## Raw Log Syncing
    async def _sync_pending_logs(self):
        if self._is_offline:
            return

        log = self.log_manager.get_next_pending_log()
        while log is not None:
            file_path = self.log_manager.get_path_to_log(log.hash)
            await self.bucket_manager.upload_log(file_path, log.hash)

            # ?: Where do we want to store the timestamp and build_id that the log was generated?
            notes = f"BuildId: {log.build_id}\nLog Captured At: {log.timestamp.strftime('%Y-%m-%d %H:%M:%S')}"
            create_file_response = await self.api.create_file(
                API.FileType.Log, log.filename, log.hash, notes
            )
            new_file_id = create_file_response["id"]

            inverter_id = await self._get_id_for_serial_number(log.serial_number)

            new_association = await self.api.create_association(
                inverter_id, new_file_id
            )

            ## Move UI row from pending to current
            row = self.view.pending_log_rows.pop(log.hash)
            self.view.show_check_icon(row, Cols.web)
            map = AssociationMapping(new_association, row)
            key = new_association["id"] + new_association["file"]["id"]

            self.associations[key] = map

            self.cache_manager.move_into_cache(
                self.log_manager.get_path_to_log(log.hash)
            )

            # Done processing. Remove pending and get next (if present)
            self.log_manager.remove_pending(log)
            log = self.log_manager.get_next_pending_log()

    async def _on_new_pending_log(self, log: PendingLog):
        inverter_id = await self._get_id_for_serial_number(log.serial_number)

        event = Event.new_raw_log(
            inverter_id,
            log.username,
            log.build_id,
            log.serial_number,
            log.filename,
            log.hash,
        )
        await self.activity_log.add(event)

        self._add_new_pending_log_row(log)
        await self._sync_pending_logs()

    def _show_pending_logs(self):
        for log in self.log_manager.get_pending_logs():
            if log.hash not in self.view.pending_log_rows:
                self._add_new_pending_log_row(log)

    def _add_new_pending_log_row(self, log: PendingLog):
        ctime = self.log_manager.stat(log.hash).st_ctime
        ctime = datetime.fromtimestamp(ctime)

        self.view.add_new_pending_log_row(log, ctime)

    def set_offline(self, is_offline):
        self.activity_syncer.set_offline(is_offline)
        self._is_offline = is_offline

        if is_offline:
            self.view.show_logged_in_status(False)
        else:
            self.view.show_logged_in_status(True, self.aws_login_manager.get_username())

        self.view.btn_sync_now.setDisabled(is_offline)

    async def _get_id_for_serial_number(self, serial_number: str):
        if serial_number not in self._inverter_id_lookup and not self._is_offline:
            inverter = await self.api.get_inverter_by_serial(serial_number)
            self._inverter_id_lookup[serial_number] = inverter["id"]

        return self._inverter_id_lookup[serial_number]

    async def debug(self):
        self.api.ws_handler.clients[0].disconnect()


class LogRenderer:
    @staticmethod
    def render_event(event: Event) -> str:
        if event.type == Event.Type.load_param_file:
            return f"Param file {event.details['filename']} loaded."  # (Hash: {event.details['fileHash'][:8]})"
        elif event.type == Event.Type.new_raw_log:
            return f"New raw log generated. ({json.dumps(event.details)})"
        elif event.type == Event.Type.param_set:
            return f"Parameter \"{event.details['paramName']}\" set to \"{event.details['paramValue']}\"."
        elif event.type == Event.Type.push_to_inverter:
            return "All settings pushed to inverter."
        else:
            return f"Unknown event type: {event.type}. Details: {event.details}"



import logging
import os
import shutil

import epyqlib.busproxy
import epyqlib.device
import epyqlib.twisted.busproxy
import epyqlib.tests.common


def assert_device_ok(device):
    attributes = (
        "name",
        "ui",
    )

    for attribute in attributes:
        assert hasattr(device, attribute)


def test_epc(customer_device_path, qtbot):
    path = os.path.splitext(customer_device_path)[0] + ".ePC"
    shutil.move(customer_device_path, path)

    device = epyqlib.device.Device(
        file=path,
        node_id=247,
    )

    assert_device_ok(device)
    device.terminate()


def test_json(customer_device_path, qtbot):
    path = os.path.splitext(customer_device_path)[0] + ".json"
    shutil.move(customer_device_path, path)

    device = epyqlib.device.Device(
        file=path,
        node_id=247,
    )

    assert_device_ok(device)
    device.terminate()


def test_epz(qtbot, zipped_customer_device_path, tmpdir):
    path = os.path.join(tmpdir, "customer.ePZ")

    shutil.move(
        zipped_customer_device_path,
        path,
    )

    device = epyqlib.device.Device(
        file=path,
        node_id=247,
    )

    assert_device_ok(device)
    device.terminate()


def test_zip(qtbot, zipped_customer_device_path, tmpdir):
    path = os.path.join(tmpdir, "customer.zIP")

    shutil.move(
        zipped_customer_device_path,
        path,
    )

    device = epyqlib.device.Device(
        file=path,
        node_id=247,
    )

    assert_device_ok(device)
    device.terminate()



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

import argparse
import csv
import logging

# import math
import signal
import sys

from PyQt5 import QtChart, QtCore, QtGui, QtWidgets

# See file COPYING in this source tree
__copyright__ = "Copyright 2017, EPC Power Corp."
__license__ = "GPLv2+"


logger = logging.getLogger(__name__)


def parse_args(args):
    parser = argparse.ArgumentParser()
    parser.add_argument("--verbose", "-v", action="count", default=0)
    parser.add_argument("--file", "-f", type=argparse.FileType("r"), required=True)

    return parser.parse_args(args)


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    args = parse_args(args=args)

    if args.verbose >= 1:
        logger.setLevel(logging.DEBUG)

    if args.verbose >= 2:
        logging.getLogger().setLevel(logging.DEBUG)

    data = read_csv(args.file.name)

    qtc(data=data)


def read_csv(filename):
    with open(filename, "r") as f:
        reader = csv.DictReader(f)

        data = {name: [] for name in reader.fieldnames}

        for row in reader:
            for k, v in row.items():
                data[k].append(float(v))

    return data


# class Chart(QtChart.QChart):
#     def mouseDoubleClickEvent(self, event):
#         if event.button() == QtCore.Qt.RightButton:
#             self.zoomReset()
#             return True
#
# class ChartView(QtChart.QChartView):
#     def mouseDoubleClickEvent(self, event):
#         if event.button() == QtCore.Qt.RightButton:
#             self.chart().zoomReset()
#             return True


class DoubleClickSignal(QtCore.QObject):
    triggered = QtCore.pyqtSignal()

    def eventFilter(self, _, event):
        if (
            isinstance(event, QtGui.QMouseEvent)
            and event.button() == QtCore.Qt.RightButton
        ):
            if event.type() == QtCore.QEvent.MouseButtonDblClick:
                self.triggered.emit()
                return True
            # elif event.type() == QtCore.QEvent.MouseButtonPress:
            # elif event.type() == QtCore.QEvent.MouseButtonRelease:

        return False


class ModifierKeySignal(QtCore.QObject):
    triggered = QtCore.pyqtSignal(int, int)
    keys = (QtCore.Qt.Key_Control, QtCore.Qt.Key_Shift, QtCore.Qt.Key_Alt)

    def eventFilter(self, _, event):
        if isinstance(event, QtGui.QKeyEvent) and event.key() in self.keys:
            self.triggered.emit(event.key(), event.type() == QtCore.QEvent.KeyPress)

        return False


def make_chart_view_zoomable(chart):
    filters = []

    f = DoubleClickSignal()
    f.triggered.connect(lambda: _zoom_reset(chart))
    filters.append(f)

    f = ModifierKeySignal()
    f.triggered.connect(lambda *args: _keyboard_modifier(chart, *args))
    filters.append(f)

    for f in filters:
        chart.installEventFilter(f)

    return filters


def _zoom_reset(chart):
    QtCore.QTimer.singleShot(0.2 * 1000, chart.chart().zoomReset)


def _keyboard_modifier(chart, key, pressed):
    mode = QtChart.QChartView.HorizontalRubberBand

    if pressed:
        if key == QtCore.Qt.Key_Shift:
            mode = QtChart.QChartView.VerticalRubberBand
        elif key == QtCore.Qt.Key_Control:
            mode = QtChart.QChartView.RectangleRubberBand

    chart.setRubberBand(mode)


class CheckableChart:
    def __init__(self):
        self.check_box = QtWidgets.QCheckBox()
        self.check_box.setCheckState(QtCore.Qt.Checked)
        self.check_box.stateChanged.connect(self._check_changed)

        self.scaling_spin_box = QtWidgets.QDoubleSpinBox()
        # TODO: set a useful width without limiting the range
        # self.scaling_spin_box.setRange(-math.inf, math.inf)
        self.scaling_spin_box.setRange(-100000, 100000)
        self.scaling_spin_box.setDecimals(3)
        self.scaling_spin_box.setValue(1)
        self.scaling_spin_box.valueChanged.connect(self.scaling_changed)

        self.chart = QtChart.QChart()
        self.series = QtChart.QLineSeries()
        self.chart.addSeries(self.series)
        self.chart.createDefaultAxes()

        # self.chart.plotAreaChanged.connect(self._plot_area_changed)
        self.original_area = None

        self.view = QtChart.QChartView(self.chart)
        self.view_filters = make_chart_view_zoomable(self.view)

        reference_height = self.check_box.minimumSizeHint().height()
        reference_height += self.scaling_spin_box.minimumSizeHint().height()
        self.view.setMinimumHeight(6 * reference_height)
        view_size_policy = self.view.sizePolicy()
        view_size_policy.setVerticalPolicy(view_size_policy.Minimum)

        self._name = None
        self.name = "<unnamed>"

        self.data = ()

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, new):
        self._name = new
        self.check_box.setText(self.name)
        self.series.setName(self.name)

    # def _plot_area_changed(self, area):
    #     if self.original_area is None:
    #         print('changed: {}'.format(area))
    #         self.original_area = area
    #     elif (area.width() > self.original_area.width()
    #             or area.height() > self.original_area.height()):
    #         print('restricted: {}'.format(area))
    #         self.chart.zoomIn(self.original_area)
    #     else:
    #         print('nop: {}'.format(area))

    def _check_changed(self, state):
        self.view.setVisible(state == QtCore.Qt.Checked)

    def set_data(self, data):
        self.data = tuple(data)
        self.update()

    def scaling_changed(self, _):
        self.update()

    def update(self):
        scale = self.scaling_spin_box.value()

        self.series.replace(
            QtGui.QPolygonF((QtCore.QPointF(x, y * scale) for x, y in self.data))
        )

        if len(self.data) > 0:
            y_minimum = min(y for _, y in self.data) * scale
            y_maximum = max(y for _, y in self.data) * scale

            delta = y_maximum - y_minimum

            if delta == 0:
                extra = 1
            else:
                extra = 0.05 * delta

            self.chart.axisY().setRange(
                y_minimum - extra,
                y_maximum + extra,
            )

            x_minimum = min(x for x, _ in self.data)
            x_maximum = max(x for x, _ in self.data)

            self.chart.axisX().setRange(
                x_minimum,
                x_maximum,
            )


class QtChartWindow(QtWidgets.QMainWindow):
    closing = QtCore.pyqtSignal()

    def __init__(self, data, parent=None):
        super().__init__(parent=parent)

        self.setWindowModality(QtCore.Qt.NonModal)

        self.central_widget = QtWidgets.QWidget()
        self.central_widget_layout = QtWidgets.QVBoxLayout()
        self.scroll_area = QtWidgets.QScrollArea()
        # http://stackoverflow.com/a/27840674/228539
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.scroll_area.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)
        self.central_widget.setLayout(self.central_widget_layout)
        self.central_widget_layout.addWidget(self.scroll_area)
        self.scroll_widget = QtWidgets.QWidget()
        self.scroll_area.setWidget(self.scroll_widget)

        self.grid_layout = QtWidgets.QGridLayout()
        # self.b1 = QtWidgets.QPushButton()
        # self.b2 = QtWidgets.QPushButton()
        # self.vlayout.addWidget()
        self.scroll_widget.setLayout(self.grid_layout)
        self.setCentralWidget(self.central_widget)

        width = QtWidgets.QApplication.desktop().screenGeometry().width()
        size = self.size()
        size.setWidth(
            QtWidgets.QDesktopWidget().availableGeometry(self).size().width() * 0.7
        )
        size.setHeight(
            QtWidgets.QDesktopWidget().availableGeometry(self).size().height() * 0.9
        )
        self.resize(size)

        # self.charts = []
        # self.views = []
        self.x_axes = []
        # self.y_axes = []
        # self.series = []
        # self.polygons = []
        self.checkable_charts = []

        for name, values in sorted(data.items()):
            if name == ".time":
                continue

            checkable_chart = CheckableChart()
            checkable_chart.name = name
            row = self.grid_layout.rowCount()

            layout = QtWidgets.QVBoxLayout()
            layout.addWidget(checkable_chart.check_box)
            layout.addWidget(checkable_chart.scaling_spin_box)
            checkable_chart.scaling_spin_box.setSizePolicy(
                QtWidgets.QSizePolicy.Fixed,
                checkable_chart.scaling_spin_box.sizePolicy().verticalPolicy(),
            )

            self.grid_layout.addLayout(layout, row, 0, 1, 1, QtCore.Qt.AlignLeft)
            self.grid_layout.addWidget(checkable_chart.view, row, 1)

            chart = checkable_chart.chart
            chart.legend().hide()
            view = checkable_chart.view

            view.setRubberBand(QtChart.QChartView.HorizontalRubberBand)
            view.setRenderHint(QtGui.QPainter.Antialiasing)

            if ".time" in data:
                d = zip(data[".time"], values)
            else:
                d = enumerate(values)
            checkable_chart.set_data(data=d)
            self.x_axes.append(chart.axisX())
            self.checkable_charts.append(checkable_chart)

        for axis in self.x_axes:
            axis.rangeChanged.connect(self.axis_range_changed)

    @QtCore.pyqtSlot("qreal", "qreal")
    def axis_range_changed(self, min, max):
        for axis in self.x_axes:
            axis.setRange(min, max)

    def closeEvent(self, event):
        self.closing.emit()


def qtc(data):
    app = QtWidgets.QApplication(sys.argv)
    window = QtChartWindow(data=data, parent=None)
    window.show()

    return app.exec()


def _entry_point():
    import traceback

    logging.basicConfig(format="%(asctime)s - %(levelname)s - %(message)s")

    def excepthook(excType, excValue, tracebackobj):
        logger.error("Uncaught exception hooked:")
        traceback.print_exception(excType, excValue, tracebackobj)

    sys.excepthook = excepthook
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    return main()


if __name__ == "__main__":
    sys.exit(_entry_point())



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

from PyQt5.QtCore import pyqtProperty
from PyQt5.QtWidgets import QWidget, QStackedLayout
from PyQt5.QtGui import QColor

import epyqlib.dualscale_ui
import epyqlib.utils.qt
import epyqlib.widgets.scale


# See file COPYING in this source tree
__copyright__ = "Copyright 2017, EPC Power Corp."
__license__ = "GPLv2+"


class DualScale(QWidget):
    def __init__(self, parent=None, in_designer=False):

        QWidget.__init__(self, parent=parent)

        self.in_designer = in_designer

        self.d_vertically_flipped = False

        self.ui = epyqlib.dualscale_ui.Ui_Form()
        self.ui.setupUi(self)

        self.scale1 = epyqlib.widgets.scale.Scale(self, in_designer)
        self.scale2 = epyqlib.widgets.scale.Scale(self, in_designer)
        # color ranges, scale markers, labels, needle painted.
        self.scale1.ui.scale.m_paintMode = 1
        # needle, cover painted.
        self.scale2.ui.scale.m_paintMode = 3
        # scale2's needle is blue
        self.scale2.ui.scale.isBlue = True
        self.stackedLayout = QStackedLayout()
        self.stackedLayout.addWidget(self.scale2)
        self.stackedLayout.addWidget(self.scale1)
        self.stackedLayout.setStackingMode(1)
        self.ui.glayout.addLayout(self.stackedLayout, 0, 0)

        # Trying to figure out how to get lower min and the higher max and change the scale
        # true_min = min(self.scale1.ui.scale.m_minimum, self.scale2.ui.scale.m_minimum)
        # true_max = max(self.scale1.ui.scale.m_maximum, self.scale2.ui.scale.m_maximum)

        # self.scale1.minimum = true_min
        # self.scale2.minimum = true_min

        # self.scale1.maximum = true_max
        # self.scale2.maximum = true_max

        # self.scale1.override_range = True
        # self.scale2.override_range = True

    @pyqtProperty("QString")
    def scale1_signal_path(self):
        return self.scale1.signal_path

    @scale1_signal_path.setter
    def scale1_signal_path(self, value):
        self.scale1.signal_path = value

    @pyqtProperty(bool)
    def scale1_label_visible(self):
        return self.scale1.label_visible

    @scale1_label_visible.setter
    def scale1_label_visible(self, new_visible):
        self.scale1.label_visible = new_visible

    @pyqtProperty("QString")
    def scale2_signal_path(self):
        return self.scale2.signal_path

    @scale2_signal_path.setter
    def scale2_signal_path(self, value):
        self.scale2.signal_path = value

    @pyqtProperty(bool)
    def scale2_label_visible(self):
        return self.scale2.label_visible

    @scale2_label_visible.setter
    def scale2_label_visible(self, new_visible):
        self.scale2.label_visible = new_visible

    @pyqtProperty(bool)
    def d_flipped(self):
        return self.d_vertically_flipped

    @d_flipped.setter
    def d_flipped(self, value):
        self.d_vertically_flipped = value
        self.scale1.s_flipped = value
        self.scale2.s_flipped = value

    @pyqtProperty(float)
    def lower_red_breakpoint(self):
        return self.scale1._breakpoints[0]

    @lower_red_breakpoint.setter
    def lower_red_breakpoint(self, breakpoint):
        self.scale1._breakpoints[0] = breakpoint
        self.scale1.update_configuration()

    @pyqtProperty(float)
    def lower_yellow_breakpoint(self):
        return self.scale1._breakpoints[1]

    @lower_yellow_breakpoint.setter
    def lower_yellow_breakpoint(self, breakpoint):
        self.scale1._breakpoints[1] = breakpoint
        self.scale1.update_configuration()

    @pyqtProperty(float)
    def upper_yellow_breakpoint(self):
        return self.scale1._breakpoints[2]

    @upper_yellow_breakpoint.setter
    def upper_yellow_breakpoint(self, breakpoint):
        self.scale1._breakpoints[2] = breakpoint
        self.scale1.update_configuration()

    @pyqtProperty(float)
    def upper_red_breakpoint(self):
        return self.scale1._breakpoints[3]

    @upper_red_breakpoint.setter
    def upper_red_breakpoint(self, breakpoint):
        self.scale1._breakpoints[3] = breakpoint
        self.scale1.update_configuration()

    @pyqtProperty(QColor)
    def lower_red_color(self):
        return self.scale1._colors[0]

    @lower_red_color.setter
    def lower_red_color(self, color):
        self.scale1._colors[0] = color
        self.scale1.update_configuration()

    @pyqtProperty(QColor)
    def lower_yellow_color(self):
        return self.scale1._colors[1]

    @lower_yellow_color.setter
    def lower_yellow_color(self, color):
        self.scale1._colors[1] = color
        self.scale1.update_configuration()

    @pyqtProperty(QColor)
    def green_color(self):
        return self.scale1._colors[2]

    @green_color.setter
    def green_color(self, color):
        self.scale1._colors[2] = color
        self.scale1.update_configuration()

    @pyqtProperty(QColor)
    def upper_yellow_color(self):
        return self.scale1._colors[3]

    @upper_yellow_color.setter
    def upper_yellow_color(self, color):
        self.scale1._colors[3] = color
        self.scale1.update_configuration()

    @pyqtProperty(QColor)
    def upper_red_color(self):
        return self.scale1._colors[4]

    @upper_red_color.setter
    def upper_red_color(self, color):
        self.scale1._colors[4] = color
        self.scale1.update_configuration()



import inspect
import pickle
import time
from collections import Callable
from datetime import datetime
from os import path
from typing import Union

import attr

from epyqlib.tabs.files.sync_config import SyncConfig


@attr.s(slots=True, auto_attribs=True)
class Event:
    class Type:
        fault_cleared = "fault-cleared"
        firmware_flashed = "firmware-flashed"
        inverter_to_nv = "inverter-to-nv"
        load_param_file = "load-param-file"
        push_to_inverter = "push-to-inverter"
        param_set = "param-set"
        new_raw_log = "new-raw-log"

    inverter_id: str
    user_id: str
    type: str
    details: dict
    timestamp: str = attr.ib(
        factory=lambda: datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")
    )

    @staticmethod
    def new_fault_cleared(inverter_id: str, user_id: str, fault_code: Union[str, int]):
        if isinstance(fault_code, int):
            fault_code = str(fault_code)

        details = {"faultCode": fault_code}
        return Event(inverter_id, user_id, Event.Type.fault_cleared, details)

    @staticmethod
    def new_firmware_flashed(inverter_id: str, user_id: str, build_id: str):
        details = {"buildId": build_id}
        return Event(inverter_id, user_id, Event.Type.firmware_flashed, details)

    @staticmethod
    def new_inverter_to_nv(inverter_id: str, user_id: str):
        return Event(inverter_id, user_id, Event.Type.inverter_to_nv, {})

    @staticmethod
    def new_param_set_event(
        inverter_id: str, user_id: str, param_name: str, param_value: str
    ):
        return Event(
            inverter_id,
            user_id,
            Event.Type.param_set,
            {"paramName": param_name, "paramValue": param_value},
        )

    @staticmethod
    def new_load_param_file(
        inverter_id: str, user_id: str, file_id: str, file_hash: str, filename: str
    ):
        details = {"fileId": file_id, "fileHash": file_hash, "filename": filename}
        return Event(inverter_id, user_id, Event.Type.load_param_file, details)

    @staticmethod
    def new_push_to_inverter(inverter_id: str, user_id: str):
        return Event(inverter_id, user_id, Event.Type.push_to_inverter, {})

    @staticmethod
    def new_raw_log(
        inverter_id: str,
        user_id: str,
        build_id: str,
        serial_number: str,
        filename: str,
        file_hash: str,
    ):
        details = {
            "buildId": build_id,
            "fileHash": file_hash,
            "filename": filename,
            "serialNumber": serial_number,
        }
        return Event(inverter_id, user_id, Event.Type.new_raw_log, details)


class ActivityLog:
    _instance = None

    def __init__(self, file_dir=None):
        file_dir = file_dir or SyncConfig.get_instance().config_dir

        self._cache_file = path.join(file_dir, "activity-cache.json")

        self._activity_cache: [Event] = []
        self._listeners = []
        self._last_write_time = 0

    @staticmethod
    def get_instance():
        if ActivityLog._instance is None:
            ActivityLog._instance = ActivityLog()
            ActivityLog._instance._read_cache_file()
        return ActivityLog._instance

    ## Managing listeners
    def register_listener(self, listener: Callable):
        self._listeners.append(listener)

    async def _notify_listeners(self, event: Event):
        coroutines = []
        for listener in self._listeners:
            result = listener(event)
            if inspect.iscoroutine(result):
                coroutines.append(result)

        for coroutine in coroutines:
            await coroutine

    ## Adding and removing events
    async def add(self, event: Event):
        self._activity_cache.append(event)

        cache_write = self._write_cache_file()

        await self._notify_listeners(event)

        await cache_write

    def remove(self, event: Event):
        self._activity_cache.remove(event)

    ## Reading events
    def has_cached_events(self):
        return len(self._activity_cache) > 0

    def read_oldest_event(self) -> Union[Event, None]:
        if not self.has_cached_events():
            return None

        return self._activity_cache[0]

    ## Cache file management
    def _read_cache_file(self):
        if path.exists(self._cache_file):
            with open(self._cache_file, "rb") as cache:
                cached_events = pickle.load(cache)
                if not isinstance(cached_events, list):
                    raise Exception(
                        f"Error reading from {self._cache_file}. Not a pickle file with a list as the root."
                    )
                self._activity_cache = cached_events + self._activity_cache

    async def _write_cache_file(self):
        now = time.time()

        if self._last_write_time + 1 > now:
            # Don't write more often than once/sec
            return

        self._last_write_time = now

        with open(self._cache_file, "wb") as file_ref:
            pickle.dump(self._activity_cache, file_ref)



#!/usr/bin/env python3

# TODO: get some docstrings in here!

# import can

# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


def referenced_files(raw_dict):
    return ()


class DeviceExtension:
    def __init__(self, device):
        self.device = device

    def post(self):
        pass

    def terminate(self):
        self.device = None


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



import logging
import can
import collections
import enum
import epyqlib.utils.twisted
import functools
import itertools
import twisted.internet.defer
import twisted.protocols.policies

from PyQt5.QtCore import QObject, pyqtSignal


logger = logging.getLogger(__name__)


def endianness_swap_2byte(b):
    # TODO: figure out the correct way to handle endianness, especially
    #       in regard to odd-length data
    # if len(b) % 2 == 1:
    #     b = b.chain(b, [0])

    return itertools.chain(
        *(
            (b, a)
            for a, b in zip(
                itertools.islice(b, 0, None, 2), itertools.islice(b, 1, None, 2)
            )
        )
    )


# http://stackoverflow.com/a/8998040/228539
def chunkit(it, n):
    it = iter(it)
    while True:
        chunk_it = itertools.islice(it, n)
        try:
            first_el = next(chunk_it)
        except StopIteration:
            return
        yield itertools.chain((first_el,), chunk_it)


class HandlerBusy(RuntimeError):
    pass


class HandlerUnknownState(RuntimeError):
    pass


class UnexpectedMessageReceived(ValueError):
    pass


class InvalidSection(ValueError):
    pass


bootloader_can_id = 0x0B081880


@enum.unique
class HandlerState(enum.Enum):
    idle = 0
    connecting = 1
    connected = 2
    disconnecting = 3
    setting_mta = 4
    downloading = 5
    download_6ing = 6
    unlocking = 7
    building_checksum = 8
    clearing_memory = 9
    uploading = 10


class Handler(QObject, twisted.protocols.policies.TimeoutMixin):
    messages_sent = pyqtSignal(int)

    def __init__(
        self,
        endianness,
        tx_id=bootloader_can_id,
        rx_id=bootloader_can_id,
        extended=True,
        parent=None,
    ):
        QObject.__init__(self, parent=parent)
        self._deferred = None
        self._stream_deferred = None
        self._internal_deferred = None
        self._chunkit = None
        self._active = False
        self._transport = None

        self._tx_id = tx_id
        self._rx_id = rx_id
        self._extended = extended

        self._send_counter = -1

        self._state = HandlerState.idle
        self._previous_state = self._state

        self._remaining_retries = 0

        self._crc = None
        self._crc_length = 0
        self.continuous_crc = None

        self._download_block_counter = 0

        self._messages_sent = 0

        self.request_memory = None

        self.endianness = endianness

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, new_state):
        logger.debug("Entering state {}".format(new_state))
        self._previous_state = self._state
        self._state = new_state

    def _new_deferred(self):
        self._deferred = twisted.internet.defer.Deferred()

    def makeConnection(self, transport):
        self._transport = transport
        logger.debug("Handler.makeConnection(): {}".format(transport))

    def connect(self, station_address=1, timeout=None):
        logger.debug("Entering connect()")
        if self._active:
            raise Exception("self._active is True")
        self._active = True

        self._new_deferred()

        if self.state is not HandlerState.idle:
            self.errback(
                HandlerBusy("Connect requested while {}".format(self.state.name))
            )
            return self._deferred

        packet = HostCommand(code=CommandCode.connect, arbitration_id=self._tx_id)
        # TODO: shouldn't be needed, just makes it agree with oz
        #       for cleaner diff
        # packet.payload[0:1] = station_address.to_bytes(2, 'little')
        packet.payload[0] = station_address
        self._send(
            packet=packet,
            state=HandlerState.connecting,
            count_towards_total=False,
            timeout=timeout,
        )

        return self._deferred

    def disconnect(self, end_of_session=0):
        logger.debug("Entering disconnect()")
        if self._active:
            raise Exception("self._active is True")
        self._active = True

        self._new_deferred()

        if self.state is not HandlerState.connected:
            self.errback(
                HandlerBusy("Disconnect requested while {}".format(self.state.name))
            )
            return self._deferred

        packet = HostCommand(code=CommandCode.disconnect, arbitration_id=self._tx_id)
        # EndOfSession rather than a temporary disconnect
        packet.payload[0] = end_of_session
        self._send(packet, state=HandlerState.disconnecting)

        logger.debug("disconnecting")
        return self._deferred

    def set_mta(self, address_extension, address):
        logger.debug("Entering set_mta()")
        if self._active:
            raise Exception("self._active is True")
        self._active = True

        self._new_deferred()

        if not isinstance(address_extension, AddressExtension):
            self.errback(
                TypeError(
                    "Expected AddressExtension, got: {}".format(type(address_extension))
                )
            )
            return self._deferred

        if self.state is not HandlerState.connected:
            self.errback(
                HandlerBusy("Set MTA requested while {}".format(self.state.name))
            )
            return self._deferred

        packet = HostCommand(code=CommandCode.set_mta, arbitration_id=self._tx_id)
        # always zero for Oz bootloader
        packet.payload[0] = 0
        packet.payload[1] = address_extension
        packet.payload[2:6] = address.to_bytes(4, self.endianness)

        self._send(packet=packet, state=HandlerState.setting_mta)

        return self._deferred

    def unlock(self, section):
        logger.debug("Entering unlock()")
        if self._active:
            raise Exception("self._active is True")
        self._active = True

        self._new_deferred()

        if not isinstance(section, Password):
            self.errback(
                InvalidSection(
                    "Invalid section password specified: {} - {}".format(
                        section.name, section.value
                    )
                )
            )
            return self._deferred

        packet = HostCommand(code=CommandCode.unlock, arbitration_id=self._tx_id)
        packet.payload[0] = 2
        packet.payload[1:3] = section.value.to_bytes(2, self.endianness)

        self._send(packet=packet, state=HandlerState.unlocking)

        return self._deferred

    def download(self, data):
        logger.debug("Entering download()")

        if self._active:
            raise Exception("self._active is True")
        self._active = True

        self._new_deferred()

        # TODO: figure out the correct way to handle endianness, especially
        #       in regard to odd-length data
        length = len(data)
        if length > 5 or length % 2 != 0:
            self.errback(TypeError("Invalid data length {}".format(length)))
            return self._deferred

        if self.state is not HandlerState.connected:
            self.errback(
                HandlerBusy("Download requested while {}".format(self.state.name))
            )
            return self._deferred

        packet = HostCommand(code=CommandCode.download, arbitration_id=self._tx_id)
        swapped_data = tuple(endianness_swap_2byte(data))
        packet.payload[0] = len(swapped_data)
        packet.payload[1 : len(swapped_data) + 1] = swapped_data

        self._send(packet=packet, state=HandlerState.downloading)

        return self._deferred

    def download_6(self, data):
        logger.debug("Entering download_6()")

        if self._active:
            raise Exception("self._active is True")
        self._active = True

        self._new_deferred()

        if len(data) != 6:
            self.errback(TypeError("Invalid data length {}".format(len(bytes))))
            return self._deferred

        if self.state is not HandlerState.connected:
            self.errback(
                HandlerBusy("Download requested while {}".format(self.state.name))
            )
            return self._deferred

        packet = HostCommand(code=CommandCode.download_6, arbitration_id=self._tx_id)
        packet.payload[:] = endianness_swap_2byte(data)

        self._send(packet=packet, state=HandlerState.download_6ing)

        return self._deferred

    # TODO: magic number 5!
    def upload(self, number_of_bytes=5, block_transfer=False):
        logger.debug("Entering upload()")

        if self._active:
            raise Exception("self._active is True")
        self._active = True

        self._new_deferred()

        # TODO: magic number 5! and 255!
        maximum = 5 if not block_transfer else 255

        if not 1 <= number_of_bytes <= maximum:
            self.errback(
                TypeError("Invalid byte count requested: {}".format(number_of_bytes))
            )
        else:
            if self.state is not HandlerState.connected:
                self.errback(
                    HandlerBusy("Upload requested while {}".format(self.state.name))
                )
            else:
                packet = HostCommand(
                    code=CommandCode.upload, arbitration_id=self._tx_id
                )
                packet.payload[0] = number_of_bytes
                self.request_memory = number_of_bytes, bytearray()

                self._send(packet=packet, state=HandlerState.uploading)

        return self._deferred

    def build_checksum(self, checksum, length):
        # length is in bytes, not addresses

        logger.debug("Entering build_checksum()")

        if self._active:
            raise Exception("self._active is True")
        self._active = True

        self._new_deferred()

        if self.state is not HandlerState.connected:
            self.errback(
                HandlerBusy("Build checksum requested while {}".format(self.state.name))
            )
            return self._deferred

        packet = HostCommand(
            code=CommandCode.build_checksum, arbitration_id=self._tx_id
        )
        logger.debug("{}, {}".format(type(length), type(checksum)))
        logger.debug(
            (length.to_bytes(4, self.endianness), checksum.to_bytes(2, self.endianness))
        )
        packet.payload[:4] = length.to_bytes(4, self.endianness)
        packet.payload[4:] = checksum.to_bytes(2, self.endianness)
        logger.debug(packet)

        self._send(packet=packet, state=HandlerState.building_checksum)

        return self._deferred

    def clear_memory(self):
        logger.debug("Entering clear_memory()")
        length = 0xFF

        if self._active:
            raise Exception("self._active is True")
        self._active = True

        self._new_deferred()

        if self.state is not HandlerState.connected:
            self.errback(
                HandlerBusy("Clear memory requested while {}".format(self.state.name))
            )
            return self._deferred

        packet = HostCommand(code=CommandCode.clear_memory, arbitration_id=self._tx_id)
        packet.payload[:4] = length.to_bytes(4, self.endianness)

        self._send(packet=packet, state=HandlerState.clearing_memory)

        return self._deferred

    def download_block(self, address_extension, address, data):
        logger.debug("Entering download_block()")
        # print('download_block(address_extension={}, address=0x{:08X})'.format(address_extension, address))
        if self._active:
            raise Exception("self._active is True")

        # self._stream_deferred = twisted.internet.defer.Deferred()

        self._chunkit = chunkit(it=data, n=6)

        # TODO: OOP this
        self._crc = None
        self._crc_length = 0
        self._download_block_counter = 0

        self._internal_deferred = self.set_mta(address_extension, address)
        self._internal_deferred.addCallback(
            lambda _, address=address, address_extension=address_extension: self._download_chunk(
                address=address, address_extension=address_extension
            )
        )
        # self._internal_deferred.addCallback(self._stream_deferred.callback)
        self._internal_deferred.addErrback(epyqlib.utils.twisted.logit)

        # return self._stream_deferred
        return self._internal_deferred

    def _download_chunk(self, address, address_extension):
        logger.debug("Entering _download_chunk()")
        try:
            chunk = next(self._chunkit)
        except StopIteration:
            chunk = ()
        else:
            chunk = tuple(chunk)

        length = len(chunk)

        download = None
        if length == 6:
            download = self.download_6
        elif length > 0:
            download = self.download

        deferred = None

        if download is not None:
            # TODO: OOP this
            self._crc = crc(data=endianness_swap_2byte(chunk), crc=self._crc)
            self.continuous_crc = crc(
                data=endianness_swap_2byte(chunk), crc=self.continuous_crc
            )
            logger.debug("Continuous CRC: {:04X}".format(self.continuous_crc))
            self._crc_length += len(chunk)

            self._download_block_counter += 1
            if self._download_block_counter >= 5:
                self._download_block_counter = 0

            deferred = download(data=chunk)
            address += length / 2
            if int(address) != address:
                # TODO: do this better or at least a unique exception
                raise Exception("ack")
            address = int(address)
            if self._download_block_counter == 0:
                deferred.addCallback(
                    lambda _, crc=self._crc, length=self._crc_length: self.build_checksum(
                        checksum=crc, length=length
                    )
                )
                self._crc = None
                self._crc_length = 0
                deferred.addCallback(
                    lambda _, address=address, address_extension=address_extension: self.set_mta(
                        address=address, address_extension=address_extension
                    )
                )
            # TODO: this just doesn't feel like good structure
            if download is not self.download:
                deferred.addCallback(
                    lambda _, address=address, address_extension=address_extension: self._download_chunk(
                        address=address, address_extension=address_extension
                    )
                )
        else:
            logger.debug("crc: {} {}".format(type(self._crc), self._crc))
            # l = lambda _: self._internal_deferred.callback(
            #         'Done downloading stream')
            if self._crc is not None:
                deferred = self.build_checksum(
                    checksum=self._crc, length=self._crc_length
                )
            #     deferred.addCallback(l)
            # else:
            #     l()

        if deferred is not None:
            deferred.addErrback(epyqlib.utils.twisted.logit)
        else:
            deferred = twisted.internet.defer.Deferred()
            deferred.callback(None)

        return deferred

        # # TODO: this smells...  really bad
        # deferred.callback('Oh so smelly...')
        # return self._internal_deferred

    @twisted.internet.defer.inlineCallbacks
    def upload_block(self, address_extension, address, octets, progress=None):
        yield self.set_mta(address=address, address_extension=address_extension)

        data = bytearray()

        remaining = octets
        update_period = octets // 100  # 1%
        since_update = 0
        while remaining > 0:
            # TODO: magic number 255!
            number_of_bytes = min(255, remaining)
            block = yield self.upload(
                number_of_bytes=number_of_bytes, block_transfer=True
            )
            remaining -= number_of_bytes

            if progress is not None:
                since_update += number_of_bytes
                if since_update >= update_period:
                    progress.update(octets - remaining)
                    since_update = 0
            data.extend(block)

        twisted.internet.defer.returnValue(data)

    def _send(self, packet, state, count_towards_total=True, timeout=None):
        if timeout is None:
            timeout = packet.command_code.timeout
        if self._send_counter < 255:
            self._send_counter += 1
        else:
            self._send_counter = 0

        packet.command_counter = self._send_counter

        self.state = state

        logger.debug("Message to be sent: {}".format(packet))
        self._transport.write(packet.message)

        if count_towards_total:
            self._messages_sent += 1
            self.messages_sent.emit(self._messages_sent)

        self.setTimeout(timeout)
        logger.debug("Timeout set to {}".format(packet.command_code.timeout))

    def dataReceived(self, msg):
        if not (
            msg.arbitration_id == self._rx_id
            and bool(msg.is_extended_id) == self._extended
        ):
            return

        if not self._active:
            return

        self.setTimeout(None)

        packet = Packet.from_message(message=msg)

        if not isinstance(packet, BootloaderReply):
            self.errback(
                UnexpectedMessageReceived("Not a bootloader reply: {}".format(packet))
            )
            return

        if self.state not in [HandlerState.connecting, HandlerState.connected]:
            if packet.command_counter != self._send_counter:
                self.errback(
                    UnexpectedMessageReceived(
                        "Reply out of sequence: expected {} but got {} - {}".format(
                            self._send_counter, packet.command_counter, packet
                        )
                    )
                )
                return

        logger.debug("packet received: {}".format(packet.command_return_code.name))

        if self.state is HandlerState.connected:
            logger.debug(
                "Unexpected message received in state connected: {}".format(packet)
            )
        elif self.state is HandlerState.connecting:
            if packet.command_return_code is not CommandStatus.acknowledge:
                self.errback(
                    UnexpectedMessageReceived(
                        "Bootloader should ack when trying to connect, instead: {}".format(
                            packet
                        )
                    )
                )
                return

            logger.debug(
                "Bootloader version: {major}.{minor}".format(
                    major=packet.payload[0], minor=packet.payload[1]
                )
            )

            dsp_code = (int(packet.payload[2]) << 8) + int(packet.payload[3])

            try:
                dsp_code = DspCode(dsp_code)
            except ValueError:
                pass
            else:
                logger.debug("DSP Part ID: {}".format(dsp_code.name))

            self.state = HandlerState.connected
            self.callback("successfully connected")
        elif self.state is HandlerState.disconnecting:
            if packet.command_return_code is not CommandStatus.acknowledge:
                self.errback(
                    UnexpectedMessageReceived(
                        "Bootloader should ack when trying to disconnect, instead: {}".format(
                            packet
                        )
                    )
                )
                return

            self.state = HandlerState.idle
            self.callback("successfully disconnected")
        elif self.state is HandlerState.setting_mta:
            if packet.command_return_code is not CommandStatus.acknowledge:
                self.errback(
                    UnexpectedMessageReceived(
                        "Bootloader should ack when trying to set MTA, instead: {}".format(
                            packet
                        )
                    )
                )
                return

            self.state = HandlerState.connected
            self.callback("successfully set MTA")
        elif self.state is HandlerState.downloading:
            if packet.command_return_code is not CommandStatus.acknowledge:
                self.errback(
                    UnexpectedMessageReceived(
                        "Bootloader should ack when trying to download, instead: {} - {}".format(
                            packet.command_return_code.name, packet
                        )
                    )
                )
                return

            self.state = HandlerState.connected
            self.callback("successfully downloaded")
        elif self.state is HandlerState.download_6ing:
            if packet.command_return_code is not CommandStatus.acknowledge:
                self.errback(
                    UnexpectedMessageReceived(
                        "Bootloader should ack when trying to download_6, instead: {}".format(
                            packet
                        )
                    )
                )
                return

            self.state = HandlerState.connected
            self.callback("successfully download_6ed")
        elif self.state is HandlerState.unlocking:
            if packet.command_return_code is not CommandStatus.acknowledge:
                self.errback(
                    UnexpectedMessageReceived(
                        "Bootloader should ack when trying to unlock, instead: {}".format(
                            packet
                        )
                    )
                )
                return

            self.state = HandlerState.connected
            self.callback("successfully unlocked {}".format(packet.payload[1]))
        elif self.state is HandlerState.building_checksum:
            if packet.command_return_code is not CommandStatus.acknowledge:
                self.errback(
                    UnexpectedMessageReceived(
                        "Bootloader should ack when trying to build checksum, instead: {}".format(
                            packet
                        )
                    )
                )
                return

            # TODO: consider verifying returned CRC data beyond just
            #       accepting the embedded side's decision to ack

            self.state = HandlerState.connected
            self.callback("successfully unlocked {}".format(packet.payload[1]))
        elif self.state is HandlerState.clearing_memory:
            if packet.command_return_code is not CommandStatus.acknowledge:
                self.errback(
                    UnexpectedMessageReceived(
                        "Bootloader should ack when trying to clear memory, instead: {} {}".format(
                            packet.command_return_code.name, packet
                        )
                    )
                )
                return

            self.state = HandlerState.connected
            self.callback("successfully unlocked {}".format(packet.payload[1]))
        elif self.state is HandlerState.uploading:
            if packet.command_return_code is not CommandStatus.acknowledge:
                self.errback(
                    UnexpectedMessageReceived(
                        "Module should ack when trying to upload, instead: {} {}".format(
                            packet.command_return_code.name, packet
                        )
                    )
                )
                return

            number_of_bytes, data = self.request_memory
            # TODO: magic number 5!
            bytes_received = min(number_of_bytes, 5)

            data.extend(packet.payload[0:bytes_received])
            number_of_bytes -= bytes_received

            if number_of_bytes == 0:
                self.state = HandlerState.connected
                self.callback(data)
            else:
                self.request_memory = number_of_bytes, data
                self.setTimeout(_command_code_properties[CommandCode.upload].timeout)
        else:
            self.errback(
                HandlerUnknownState("Handler in unknown state: {}".format(self.state))
            )
            return

    def timeoutConnection(self):
        message = "Handler timed out while in state: {}".format(self.state)
        logger.debug(message)
        self._active = False
        if self._previous_state in [HandlerState.idle]:
            self.state = self._previous_state
        self._deferred.errback(epyqlib.utils.twisted.RequestTimeoutError(message))

    def callback(self, payload):
        self._active = False
        logger.debug("calling back for {}".format(self._deferred))
        self._deferred.callback(payload)

    def errback(self, payload):
        self._active = False
        logger.debug("erring back for {}".format(self._deferred))
        logger.debug("with payload {}".format(payload))
        self._deferred.errback(payload)

    def cancel(self):
        self._active = False
        self.setTimeout(None)
        self._deferred.cancel()


def crc(data, crc=None):
    if crc is None:
        crc = 0xFFFF

    for byte in data:
        crc ^= byte & 0x00FF

        for _ in range(8):
            if (crc & 0x0001) != 0:
                crc = (crc >> 1) ^ 0xA001
            else:
                crc = crc >> 1

    return crc


class IdentifierTypeError(ValueError):
    pass


class MessageLengthError(ValueError):
    pass


class WindowSlice:
    def __init__(self, original, offset):
        self._original = original
        self._offset = offset

    def _offsetit(self, key):
        if isinstance(key, slice):
            start = 0 if key.start is None else key.start
            start += self._offset
            stop = None if key.stop is None else key.stop + self._offset

            key = slice(start, stop, key.step)
        else:
            key += self._offset

        return key

    def __getitem__(self, key):
        key = self._offsetit(key)
        return self._original.__getitem__(key)

    def __setitem__(self, key, value):
        key = self._offsetit(key)
        return self._original.__setitem__(key, value)

    def __str__(self):
        return str(self._original[self._offsetit(slice(None))])


class Packet:
    def __init__(
        self,
        counter_index,
        payload_start,
        extended_id=True,
        arbitration_id=bootloader_can_id,
        dlc=8,
        *args,
        **kwargs,
    ):
        # TODO: avoid repetition of required values
        # TODO: block changing of required values
        if not extended_id:
            raise IdentifierTypeError("Identifier type must be set to extended")

        if dlc != 8:
            raise MessageLengthError("Message length must be 8 but is {}".format(dlc))

        kwargs.setdefault("data", [0] * dlc)

        self.message = can.Message(
            is_extended_id=extended_id,
            arbitration_id=arbitration_id,
            dlc=dlc,
            *args,
            **kwargs,
        )

        self._counter_index = counter_index
        self._payload = WindowSlice(self.message.data, payload_start)

    @property
    def payload(self):
        return self._payload

    @classmethod
    def from_message(cls, message):
        if message.data[0] == 0xFF:
            c = BootloaderReply
        else:
            c = HostCommand

        return c(
            code=None,
            timestamp=message.timestamp,
            is_remote_frame=message.is_remote_frame,
            extended_id=message.is_extended_id,
            is_error_frame=message.is_error_frame,
            arbitration_id=message.arbitration_id,
            dlc=message.dlc,
            data=message.data,
        )

    @property
    def command_counter(self):
        return self.message.data[self._counter_index]

    @command_counter.setter
    def command_counter(self, counter):
        self.message.data[self._counter_index] = int(counter)


class HostCommand(Packet):
    def __init__(self, code, *args, **kwargs):
        super().__init__(counter_index=1, payload_start=2, *args, **kwargs)

        if code is not None:
            self.command_code = code

    @property
    def command_code(self):
        return CommandCode(self.message.data[0])

    @command_code.setter
    def command_code(self, code):
        self.message.data[0] = int(code)


class BootloaderReply(Packet):
    def __init__(self, code, *args, **kwargs):
        super().__init__(counter_index=2, payload_start=3, *args, **kwargs)

        if code is not None:
            self.command_return_code = code

    @property
    def command_return_code(self):
        return CommandStatus(self.message.data[1])

    @command_return_code.setter
    def command_return_code(self, code):
        self.message.data[1] = int(code)


@enum.unique
class CommandCode(enum.IntEnum):
    connect = 0x01
    set_mta = 0x02
    download = 0x03
    upload = 0x04
    disconnect = 0x07
    build_checksum = 0x0E
    clear_memory = 0x10
    unlock = 0x13
    action_service = 0x21
    download_6 = 0x23

    # def __getattr__(self, name):
    #     return getattr(_command_code_properties[self], name)

    @property
    def timeout(self):
        return _command_code_properties[self].timeout


CommandCodeProperties = collections.namedtuple("CommandCodeProperties", ["timeout"])


_command_code_properties = {
    CommandCode.connect: CommandCodeProperties(timeout=5),
    CommandCode.set_mta: CommandCodeProperties(timeout=5),
    CommandCode.download: CommandCodeProperties(timeout=5),
    CommandCode.disconnect: CommandCodeProperties(timeout=5),
    CommandCode.build_checksum: CommandCodeProperties(timeout=5),
    CommandCode.clear_memory: CommandCodeProperties(timeout=30),
    CommandCode.unlock: CommandCodeProperties(timeout=5),
    CommandCode.action_service: CommandCodeProperties(timeout=5),
    CommandCode.download_6: CommandCodeProperties(timeout=5),
    CommandCode.upload: CommandCodeProperties(timeout=5),
}


@enum.unique
class Password(enum.IntEnum):
    dsp_flash = 0x1234
    # eeprom_bootloader = 0xABCD
    # eeprom_data = 0x9876


@enum.unique
class Password(enum.IntEnum):
    dsp_flash = 0x1234
    # eeprom_bootloader = 0xABCD
    # eeprom_data = 0x9876


@enum.unique
class DspCode(enum.IntEnum):
    _2801 = 0x002C
    _2802 = 0x0024
    _2806 = 0x0034
    _2808 = 0x003C
    _2809 = 0x00FE
    _28232 = 0x00E6
    _28234 = 0x00E7
    _28235 = 0x00E8
    _28332 = 0x00ED
    _28334 = 0x00EE
    _28335 = 0x00EF
    _28069PZP = 0x009E
    _28069UPZ = 0x009F
    _28069PFP = 0x009C
    _28069UPFP = 0x009D


# TODO: separate bootloader vs. embedded enumerations
# @enum.unique
class AddressExtension(enum.IntEnum):
    flash_memory = 0x00
    configuration_registers = 0x01
    eeprom_memory_bootloader = 0x02
    eeprom_memory_data = 0x03

    raw = 0x00
    data_logger = 0x01


@enum.unique
class ErrorCategory(enum.Enum):
    timeout = -1
    c0 = 0
    c1 = 1
    c2 = 2
    c3 = 3

    @property
    def description(self):
        return _error_category_properties[self].description

    @property
    def action(self):
        return _error_category_properties[self].action

    @property
    def retries(self):
        return _error_category_properties[self].retries


ErrorCategoryProperties = collections.namedtuple(
    "ErrorCategoryProperties", ["description", "action", "retries"]
)


_error_category_properties = {
    ErrorCategory.timeout: ErrorCategoryProperties(
        description="No handshake message", action="retry", retries=2
    ),
    ErrorCategory.c0: ErrorCategoryProperties(
        description="Warning", action=None, retries=None
    ),
    ErrorCategory.c1: ErrorCategoryProperties(
        description="Spurious (comm. Error, busy, ..)",
        action="Wait (ACK or timeout)",
        retries=2,
    ),
    ErrorCategory.c2: ErrorCategoryProperties(
        description="Resolvable (temp, power loss, ..)",
        action="reinitialize",
        retries=1,
    ),
    ErrorCategory.c3: ErrorCategoryProperties(
        description="Unresolvable (setup, overload, ..)",
        action="terminate",
        retries=None,
    ),
}


@enum.unique
class CommandStatus(enum.IntEnum):
    acknowledge = 0x00  # no error
    processor_busy = 0x10
    unknown_command = 0x30
    command_syntax = 0x31
    parameters_out_of_range = 0x32
    access_denied = 0x33
    access_locked = 0x35
    resource_function_unavailable = 0x36
    operational_failure = 0x7F

    @property
    def error_category(self):
        return _command_status_properties[self].error_category

    @property
    def state_transition_to(self):
        return _command_status_properties[self].state_transition_to


CommandStatusProperties = collections.namedtuple(
    "CommandStatusProperties", ["error_category", "state_transition_to"]
)


_command_status_properties = {
    CommandStatus.acknowledge: CommandStatusProperties(
        error_category=None, state_transition_to=None
    ),
    CommandStatus.processor_busy: CommandStatusProperties(
        error_category=ErrorCategory.c1, state_transition_to="Fault"
    ),
    CommandStatus.unknown_command: CommandStatusProperties(
        error_category=ErrorCategory.c3, state_transition_to="Fault"
    ),
    CommandStatus.command_syntax: CommandStatusProperties(
        error_category=ErrorCategory.c3, state_transition_to="Fault"
    ),
    CommandStatus.parameters_out_of_range: CommandStatusProperties(
        error_category=ErrorCategory.c3, state_transition_to="Fault"
    ),
    CommandStatus.access_denied: CommandStatusProperties(
        error_category=ErrorCategory.c3, state_transition_to="Fault"
    ),
    CommandStatus.access_locked: CommandStatusProperties(
        error_category=ErrorCategory.c3, state_transition_to="Fault"
    ),
    CommandStatus.resource_function_unavailable: CommandStatusProperties(
        error_category=ErrorCategory.c3, state_transition_to="Fault"
    ),
    CommandStatus.operational_failure: CommandStatusProperties(
        error_category=ErrorCategory.c3, state_transition_to="Fault"
    ),
}

if __name__ == "__main__":
    import sys
    import traceback

    def excepthook(excType, excValue, tracebackobj):
        logger.debug("Uncaught exception hooked:")
        traceback.print_exception(excType, excValue, tracebackobj)

    sys.excepthook = excepthook
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    sys.exit(main())



import collections
import os
import math
import pathlib

import attr
import click
import epyqlib.pm.valuesetmodel
import epyqlib.attrsmodel
import graham
import marshmallow


def create_path_attribute():
    return attr.ib(
        converter=pathlib.Path,
        metadata=graham.create_metadata(
            field=marshmallow.fields.String(),
        ),
    )


def to_list_of_pathlib(l):
    return [pathlib.Path(path) for path in l]


@graham.schemify(tag="valueset_overlay_recipe")
@attr.s
class OverlayRecipe:
    output_path = create_path_attribute()
    base_pmvs_path = create_path_attribute()
    overlay_pmvs_paths = attr.ib(
        converter=to_list_of_pathlib,
        metadata=graham.create_metadata(
            field=marshmallow.fields.List(marshmallow.fields.String()),
        ),
    )


@graham.schemify(tag="valueset_overlay_configuration")
@attr.s
class OverlayConfiguration:
    output_path = create_path_attribute()
    recipes = attr.ib(
        metadata=graham.create_metadata(
            field=marshmallow.fields.List(
                marshmallow.fields.Nested(graham.schema(OverlayRecipe))
            ),
        ),
    )
    reference_path = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_pathlib_or_none,
    )
    path = attr.ib(default=None)

    @classmethod
    def load(cls, path):
        schema = graham.schema(cls)

        configuration_text = path.read_text(encoding="utf-8")
        configuration = schema.loads(configuration_text).data

        configuration.path = path
        if configuration.reference_path is None:
            configuration.reference_path = path.parent

        return configuration

    def recipe_output_path(self, recipe):
        return self.reference_path / self.output_path / recipe.output_path

    def recipe_base_pmvs_path(self, recipe):
        return self.reference_path / recipe.base_pmvs_path

    def recipe_overlay_pmvs_paths(self, recipe):
        return [self.reference_path / path for path in recipe.overlay_pmvs_paths]

    def raw(self, echo=lambda *args, **kwargs: None):
        input_modification_times = []
        output_modification_times = []

        input_modification_times.append(self.path.stat().st_mtime)

        for recipe in self.recipes:
            output_path = self.recipe_output_path(recipe=recipe)
            echo(f"Checking: {os.fspath(output_path)}")

            try:
                stat = output_path.stat()
            except FileNotFoundError:
                output_modification_time = -math.inf
            else:
                output_modification_time = stat.st_mtime
            output_modification_times.append(output_modification_time)

            overlay_pmvs_paths = self.recipe_overlay_pmvs_paths(recipe=recipe)

            for overlay_pmvs_path in overlay_pmvs_paths:
                input_modification_times.append(
                    overlay_pmvs_path.stat().st_mtime,
                )

        return min(output_modification_times) < max(input_modification_times)


@click.group(name="value-sets")
def group():
    pass


@group.group()
def recipes():
    pass


@recipes.command(name="generate")
@click.option(
    "--input",
    "value_set_path_strings",
    type=click.Path(dir_okay=False, readable=True, resolve_path=True),
    multiple=True,
    required=True,
)
@click.option(
    "--common-output",
    "common_output_path_string",
    type=click.Path(dir_okay=False, readable=True, resolve_path=True),
    required=True,
)
def generate_recipes(value_set_path_strings, common_output_path_string):
    value_set_paths = [
        pathlib.Path(path_string) for path_string in value_set_path_strings
    ]
    common_output_path = pathlib.Path(common_output_path_string)

    common_value_set = epyqlib.pm.valuesetmodel.create_blank()

    value_sets = [
        epyqlib.pm.valuesetmodel.loadp(value_set_path)
        for value_set_path in value_set_paths
    ]

    all_parameters_by_uuid = collections.defaultdict(list)
    for value_set in value_sets:
        for parameter in value_set.model.root.children:
            all_parameters_by_uuid[parameter.parameter_uuid].append(parameter)

    common_uuids = [
        uuid_
        for uuid_, parameters in all_parameters_by_uuid.items()
        if len(parameters) == len(value_sets)
    ]

    common_and_equal_uuids = [
        uuid_
        for uuid_ in common_uuids
        if 1 == len({parameter.value for parameter in all_parameters_by_uuid[uuid_]})
    ]

    for uuid_ in common_and_equal_uuids:
        reference_parameter = all_parameters_by_uuid[uuid_][0]

        new_parameter = epyqlib.pm.valuesetmodel.Parameter(
            name=reference_parameter.name,
            value=reference_parameter.value,
            parameter_uuid=reference_parameter.parameter_uuid,
            readable=reference_parameter.readable,
            writable=reference_parameter.writable,
        )

        common_value_set.model.root.append_child(new_parameter)

    for value_set in value_sets:
        value_set.strip_common(reference=common_value_set)

    common_value_set.save(path=common_output_path)
    for value_set in value_sets:
        new_name = common_output_path.stem + "-" + value_set.path.name
        path = common_output_path.parent / new_name
        value_set.save(path=path)


@recipes.command(name="cook")
@click.option(
    "--configuration",
    "configuration_path_string",
    type=click.Path(dir_okay=False, readable=True, resolve_path=True),
)
@click.option(
    "--if-raw/--assume-raw",
    "only_if_raw",
    default=False,
)
def cli(configuration_path_string, only_if_raw):
    configuration_path = pathlib.Path(configuration_path_string)

    configuration = OverlayConfiguration.load(configuration_path)

    if only_if_raw:
        if not configuration.raw(echo=click.echo):
            click.echo(
                "Generated files appear to be up to date, skipping cooking",
            )

            return

        click.echo("Generated files appear to be out of date, starting cooking")

    for recipe in configuration.recipes:
        output_path = configuration.recipe_output_path(recipe=recipe)
        click.echo(f"Creating: {os.fspath(output_path)}")

        base_pmvs_path = configuration.recipe_base_pmvs_path(recipe=recipe)

        overlay_pmvs_paths = configuration.recipe_overlay_pmvs_paths(
            recipe=recipe,
        )

        all_pmvs_paths = [base_pmvs_path, *overlay_pmvs_paths]

        click.echo(
            "\n".join(f"    {os.fspath(path)}" for path in all_pmvs_paths),
        )

        result_value_set = epyqlib.pm.valuesetmodel.loadp(base_pmvs_path)

        for path in overlay_pmvs_paths:
            overlay_value_set = epyqlib.pm.valuesetmodel.loadp(path)
            result_value_set.overlay(overlay_value_set)

        output_path.parent.mkdir(parents=True, exist_ok=True)
        result_value_set.save(path=output_path)



import asyncio
import json
from enum import Enum

import treq
from epyqlib.tabs.files.activity_log import Event
from epyqlib.tabs.files.websocket_handler import WebSocketHandler
import twisted.internet
from twisted.internet.defer import ensureDeferred
from twisted.python.failure import Failure
from twisted.web.iweb import IResponse
from typing import Callable, Dict, List, Coroutine


class GraphQLException(Exception):
    pass


class InverterNotFoundException(Exception):
    pass


class API:
    _tag = "[Graphql API]"

    list_inverters_query = """
        query {
            listInverters(limit: 200) {
                items {
                    id,
                    serialNumber,
                    notes
                }
                nextToken
            }
        }
    """

    list_inverters = {"query": list_inverters_query, "variables": {}}

    _all_inverter_fields = """
                id 
                
                deploymentDate 
                manufactureDate 
                model { name partNumber id __typename } 
                notes 
                serialNumber 
                site { 
                    name 
                    id 
                    customer { name id __typename } 
                    __typename 
                } 
                testDate 
                __typename 
    """

    get_inverter_string = (
        """
        query ($inverterId: ID!) { 
            getInverter(id: $inverterId) {"""
        + _all_inverter_fields
        + """
            }
        } 
    """
    )

    def _get_inverter_query(self, inverter_id: str):
        return {
            "query": self.get_inverter_string,
            "variables": {"inverterId": inverter_id},
        }

    _get_inverter_by_sn_string = (
        """
        query ($serialNumber: String!) { 
            getInverterBySN(serialNumber: $serialNumber) {"""
        + _all_inverter_fields
        + """
            }
        } 
    """
    )

    def _get_inverter_by_sn_query(self, serial_number: str):
        return {
            "query": self._get_inverter_by_sn_string,
            "variables": {"serialNumber": serial_number},
        }

    _frag_associations_field = """
            fragment associationFields on Association {
                id
                customer { name }
                file {id, createdBy, createdAt, description, filename, hash, notes, owner, type, version}
                model {name}
                inverter {id, serialNumber}
                site {name}

                # updatedAt
            }
        """

    _get_association_str = (
        _frag_associations_field
        + """
                query ($id: ID!, $fileId: ID!) {
                    getAssociation(id: $id, fileId: $fileId) {
                       ...associationFields
                    }
                }
            """
    )

    def _get_association_query(self, association_id: str, file_id: str):
        return {
            "query": self._get_association_str,
            "variables": {"id": association_id, "fileId": file_id},
        }

    _get_associations_str = (
        _frag_associations_field
        + """
                query ($serialNumber: String) {
                    getInverterAssociations(serialNumber: $serialNumber) {
                        items {
                           ...associationFields
                        }
                    }
                }
            """
    )

    def _get_associations_query(self, serial_number: str):
        return {
            "query": self._get_associations_str,
            "variables": {"serialNumber": serial_number},
        }

    _get_associations_for_customer_str = (
        _frag_associations_field
        + """
           query {
                getAssociationsForCustomer {
                    serialNumber,
                    associations {
                        ...associationFields
                    }
                }
            }
        """
    )

    def _get_association_for_customer_query(self):
        return {"query": self._get_associations_for_customer_str, "variables": {}}

    _create_activity = """
        mutation Name (
            $inverterId: String!,
            $timestamp: String!,
            $type: String!,
            $details: AWSJSON
        ){
            createActivity(
                inverterId: $inverterId,
                timestamp: $timestamp,
                type: $type,
                details: $details
            ) {
                  inverterId
                  details
                  timestamp
                  type
                  
                  createdBy
            }
        }
    """

    def _get_create_activity_mutation(
        self, inverter_id: str, timestamp: str, type: str, details: dict
    ):
        return {
            "query": self._create_activity,
            "variables": {
                "details": json.dumps(details),
                "inverterId": inverter_id,
                "timestamp": timestamp,
                "type": type,
            },
        }

    _frag_files_fields = """
        fragment fileFields on File {
                id
                createdAt
                createdBy
                updatedAt
                updatedBy
                association {id}
                description
                filename
                hash
                notes
                owner
                version
                type
        }
    """

    _create_file_mutation = (
        _frag_files_fields
        + """
        mutation CreateFile (
            $filename: String!,
            $hash: String!,
            $notes: String,
            $type: FileType
            
        ){
            createFile(
                filename: $filename,
                hash: $hash,
                notes: $notes,
                type: $type 
            ) {
                ...fileFields
            }
        }
    """
    )

    class FileType(Enum):
        Firmware = "Firmware"
        Log = "Log"
        Other = "Other"
        Parameter = "Parameter"
        PMVS = "PMVS"

    def _get_create_file_mutation(
        self, type: FileType, filename: str, hash: str, notes: str
    ):
        return {
            "query": self._create_file_mutation,
            "variables": {
                "type": type.value,
                "filename": filename,
                "hash": hash,
                "notes": notes,
            },
        }

    _create_association_mutation = """
        mutation (
            $fileId: ID!,
            $inverterId: ID
        ){
            createAssociation (
                fileId: $fileId,
                inverterId: $inverterId
            ) {
                id
                file { id filename notes hash createdAt }
                inverter { id, serialNumber }
            }
        }
    """

    def _get_create_association_mutation(self, inverterId: str, fileId: str):
        return {
            "query": self._create_association_mutation,
            "variables": {"fileId": fileId, "inverterId": inverterId},
        }

    _update_file_notes = """        
        mutation ($fileId: ID!, $description: String, $notes: String) {
            updateFile(id: $fileId, description: $description, notes: $notes) {
                id
                hash
                description
                notes
            }
        }
    """

    def _get_update_file_notes_mutation(
        self, file_id: str, description: str, notes: str
    ):
        return {
            "query": self._update_file_notes,
            "variables": {
                "description": description,
                "fileId": file_id,
                "notes": notes,
            },
        }

    urls = {
        "internal": "https://kew33ruufvh5bpocxosmuegbvq.appsync-api.us-west-2.amazonaws.com/graphql",
        "client": "https://nd6i76lygfaoxoe4xxrxnwh5ty.appsync-api.us-west-2.amazonaws.com/graphql",
    }

    def __init__(self, environment: str):
        self.ws_handler = WebSocketHandler()
        self.server_url = self.urls[environment]
        self.headers = {}

    async def _make_request(self, body):
        try:
            response: IResponse = await treq.post(
                self.server_url, headers=(self.headers), json=body
            )
            if response.code >= 400:
                raise GraphQLException(
                    [f"{response.code} {response.phrase.decode('ascii')}"]
                )
        except Exception as e:
            print(
                f"{self._tag} Error during outgoing query: {body['query']} {body['variables']}"
            )
            raise e

        # content = await treq.content(response)
        response = await treq.json_content(response)

        if response.get("errors") is not None:
            print(f"{self._tag} Errors encountered making graphQL request:")
            print(f"{self._tag} Outgoing query: {body['query']}")
            print(
                f"{self._tag} Outgoing variables: {json.dumps(body['variables'], indent=2)}"
            )
            for error in response["errors"]:
                print(f"{self._tag} Error encountered: {json.dumps(error, indent=2)}")
            messages = [
                f"{e.get('errorType') or 'Error'}: {e['message']}"
                for e in response["errors"]
            ]

            raise GraphQLException(messages)

        return response

    async def fetch_inverter_list(self):
        response = await self._make_request(self.list_inverters)

        return response["data"]["listInverters"]["items"]

    async def get_inverter(self, inverter_id: str):
        response = await self._make_request(self._get_inverter_query(inverter_id))
        return response["data"]["getInverter"]

    async def get_inverter_by_serial(self, serial_number: str):
        response = await self._make_request(
            self._get_inverter_by_sn_query(serial_number)
        )
        return response["data"]["getInverterBySN"]

    async def get_association(self, association_id: str, file_id: str):
        response = await self._make_request(
            self._get_association_query(association_id, file_id)
        )
        return response["data"]["getAssociation"]

    async def get_associations(self, serial_number: str):
        """
        :raises InverterNotFoundException
        """
        try:
            response = await self._make_request(
                self._get_associations_query(serial_number)
            )
            return response["data"]["getInverterAssociations"]["items"]
        except GraphQLException as e:
            args: List[str] = e.args
            for message in args:
                if "Unable to find inverter" in message:
                    raise InverterNotFoundException(message)
            raise e

    async def get_associations_for_customer(self) -> Dict[str, Dict]:
        """
        :return: Dict of serialNumber -> association
        """
        response = await self._make_request(self._get_association_for_customer_query())
        items = response["data"]["getAssociationsForCustomer"]
        return {item["serialNumber"]: item["associations"] for item in items}

    async def create_activity(self, event: Event):
        # details_json = json.dumps(event.details)
        request_body = self._get_create_activity_mutation(
            event.inverter_id, event.timestamp, event.type, event.details
        )
        print("[Graphql] Sending create activity request: " + json.dumps(request_body))
        response = await self._make_request(request_body)
        print(json.dumps(response))

    async def create_file(
        self, type: FileType, filename: str, hash: str, notes: str = None
    ):
        response = await self._make_request(
            self._get_create_file_mutation(type, filename, hash, notes)
        )
        return response["data"]["createFile"]

    async def create_association(self, inverterId: str, fileId: str):
        response = await self._make_request(
            self._get_create_association_mutation(inverterId, fileId)
        )
        return response["data"]["createAssociation"]

    async def test_connection(self):
        response = await self._make_request({"query": "{ __typename}"})

    async def set_file_notes(self, file_id: str, description: str, notes: str):
        response = await self._make_request(
            self._get_update_file_notes_mutation(file_id, description, notes)
        )
        return response["data"]["updateFile"]

    async def subscribe(
        self,
        customer_id: str,
        on_message: Callable[[str, dict], Coroutine],
        on_close: Callable[[], Coroutine] = None,
    ):
        print(
            f"{self._tag} Subscribing to events for public events and events for context {customer_id}"
        )

        gql = """subscription($customerId: String!) {
                    # orgFileCreated: fileCreated(owner: $customerId) { id }
                    orgFileUpdated: fileUpdated(owner: $customerId) { id }
                    orgFileDeleted: fileDeleted(owner: $customerId) { id }
                    
                    # publicFileCreated: fileCreated(owner: "public") { id }
                    publicFileUpdated: fileUpdated(owner: "public") { id }
                    publicFileDeleted: fileDeleted(owner: "public") { id }
                    
                    orgAssociationCreated: associationCreated(owner: $customerId) { id }
                    orgAssociationDeleted: associationDeleted(owner: $customerId) { id }
                    
                    publicAssociationCreated: associationCreated(owner: "public") { id }
                    publicAssociationDeleted: associationDeleted(owner: "public") { id }
                }"""

        query = {"query": gql, "variables": {"customerId": customer_id}}

        response = await self._make_request(query)
        self.ws_handler.connect(response, on_message, on_close)

    def is_subscribed(self):
        return self.ws_handler.is_subscribed()

    async def unsubscribe(self):
        if self.ws_handler.is_subscribed():
            self.ws_handler.resubscribe = False
            self.ws_handler.disconnect()

    def set_id_token(self, id_token: str):
        self.headers["Authorization"] = id_token


def main(reactor):
    from twisted.internet.defer import ensureDeferred

    api = API()
    deferred = ensureDeferred(api.get_associations("TestInv"))
    deferred.addCallback(succ)
    deferred.addErrback(err)
    return deferred


def succ(body):
    print("SUCCESS")
    print(body)


def err(error: Failure):
    print("ERROR ENCOUNTERED")
    print(error.type)
    print(error.value)
    print(error.getBriefTraceback())
    twisted.internet.reactor.stop()


if __name__ == "__main__":
    # from twisted.internet.task import react
    # react(main)

    api = API()
    # d = ensureDeferred(api.subscribe())
    api.set_id_token(
        "eyJraWQiOiJweldEOXl5WFdNOW82MGdLWVMxREdXZWFGc2lNcWNGM3BcL1ZTZnNuVU5ZVT0iLCJhbGciOiJSUzI1NiJ9.eyJzdWIiOiIzZTA2OGQ4Yi01NTJmLTQ4YmYtYjVhZi1kZGViMDlhOTc3OWMiLCJjb2duaXRvOmdyb3VwcyI6WyJlcGMiXSwiZW1haWxfdmVyaWZpZWQiOnRydWUsImlzcyI6Imh0dHBzOlwvXC9jb2duaXRvLWlkcC51cy13ZXN0LTIuYW1hem9uYXdzLmNvbVwvdXMtd2VzdC0yXzhyelNSRFBHNiIsInBob25lX251bWJlcl92ZXJpZmllZCI6dHJ1ZSwiY29nbml0bzp1c2VybmFtZSI6InRlc3RlciIsImNvZ25pdG86cm9sZXMiOlsiYXJuOmF3czppYW06OjY3NDQ3NTI1NTY2Njpyb2xlXC9Db2duaXRvX2VwY0F1dGhfUm9sZSJdLCJhdWQiOiI0MTZncTRtZHBvczU1Y2ppcjFoNXU4ZzNzbCIsImV2ZW50X2lkIjoiYjE4ZDViYjUtMzZjMS0xMWU5LTk0NTItM2Q4MzQyOGExYjBjIiwidG9rZW5fdXNlIjoiaWQiLCJhdXRoX3RpbWUiOjE1NTA4NTQxMTQsInBob25lX251bWJlciI6IisxMTExMjIyMzM0NCIsImV4cCI6MTU1MTgzNDgzNSwiaWF0IjoxNTUxODMxMjM1LCJlbWFpbCI6ImJlbi5iZXJyeUBjcm9zc2NvbW0uY29tIn0.PZvs44g9BnAH5y34WSgYycZ2qvlrq2H57YSeKj2TrkTN53zJyuVET6YwSkb_Jvsq6QtQzLeAIymg5_jSr0o5cYOu9dp1rTgqQ1KdbPsPcOWPV9u7jilmPK-LeErJHP9zmlcpSvgurnJWet3tKgJhrIz2DD9nU6JINf1MvsINyIuBx8D3stivBPR3uAjO_cCwdwwmBa88txa35XLWPRMNReCnMTfxBcdT4tOBL_e1Wg9Cf5qUmxpagB9rYn3foOG5tjzRqb2XQ9v0UadPXzQYEKFs4dbThR2O5jrWQgvGRWH73TYS7L2euSlbyge9PEL6UapmhZ0Y0DyHilEEb8tZIg"
    )
    # d = ensureDeferred(api.create_file(API.FileType.Log, "testlog.log", "testhash"))
    # d = ensureDeferred(api.create_association("TestInv", "TestFile"))
    # d = ensureDeferred(api.set_file_notes("a5ef9c19-6592-47a7-ab2d-f3c7bafef51c", "Python notes"))
    # d = ensureDeferred(api.get_associations("0"))
    d = ensureDeferred(api.get)
    d.addCallback(succ)
    d.addErrback(err)
    # ensureDeferred(api.test_connection())
    twisted.internet.reactor.run()



import threading

import psutil


def spy(*ignore):
    """
    Make sure to ignore callables that are more than just their __call__
    because this will destroy everything else.
    """

    def inner(cls):
        class Spy(cls):
            def __getattribute__(self, name):
                attribute = super().__getattribute__(name)
                if hasattr(attribute, "__call__") and name not in ignore:

                    def mark(self, marker):
                        print(
                            "{marker} "
                            ">Thread Id 0x{thread_id:012x}< "
                            "{repr} "
                            "{marker} "
                            "{name}()".format(
                                marker=marker,
                                repr="<{cls} object at 0x{id:012x}>".format(
                                    cls=type(self).__bases__[0].__name__, id=id(self)
                                ),
                                name=name,
                                thread_id=threading.get_ident(),
                            )
                        )

                    def report(*args, **kwargs):
                        mark(self, " +")
                        result = attribute(*args, **kwargs)
                        mark(self, "- ")

                        return result

                    return report
                else:
                    return attribute

        return Spy

    return inner


class ReportThreads:
    def __init__(self):
        self.process = psutil.Process()
        self.i = 0
        self.previous_threads = set()

    def __call__(self, description=""):
        current_threads = {t.id for t in self.process.threads()}

        started_threads = current_threads - self.previous_threads
        stopped_threads = self.previous_threads - current_threads

        print("report_threads()", str(self.i).rjust(5), description)
        print("    ", threading.enumerate())

        x = (
            ("current", current_threads),
            ("started", started_threads),
            ("stopped", stopped_threads),
        )

        for d, v in x:
            print("    ", d, sorted(v))

        self.i += 1
        self.previous_threads = current_threads


report_threads = ReportThreads()



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

import functools
import epyqlib.utils.qt
from PyQt5 import QtGui, QtDesigner
import os
from PyQt5.QtCore import QFileInfo, qDebug
import sys

# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


class AbstractPlugin(QtDesigner.QPyDesignerCustomWidgetPlugin):
    # https://wiki.python.org/moin/PyQt/Using_Python_Custom_Widgets_in_Qt_Designer

    def __init__(self, parent=None, in_designer=True):
        QtDesigner.QPyDesignerCustomWidgetPlugin.__init__(self, parent=parent)

        self.in_designer = in_designer

        self._group = "EPC - Signals"
        self._icon = os.path.join(
            QFileInfo.absolutePath(QFileInfo(__file__)), "icon.ico"
        )
        self._init = None
        self._module_path = None
        self._name = None
        self._tooltip = None
        self._whats_this = None

        self.initialized = False

    def initialize(self, core):
        if self.initialized:
            return

        self.initialized = True

    def isInitialized(self):
        return self.initialized

    def createWidget(self, parent):
        return self._init(parent, in_designer=True)

    def name(self):
        return self._name

    def group(self):
        return self._group

    def icon(self):
        if self._icon is not None:
            return QtGui.QIcon(self._icon)
        else:
            QtDesigner.QPyDesignerCustomWidgetPlugin.icon(self)

    def toolTip(self):
        return self._tooltip

    def whatsThis(self):
        return self._whats_this

    def isContainer(self):
        return False

    def includeFile(self):
        return self._module_path


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

import can
import epyqlib.pyqabstractitemmodel
import functools
import logging
import sys
import time

from collections import OrderedDict
from epyqlib.abstractcolumns import AbstractColumns
import epyqlib.device
from epyqlib.treenode import TreeNode
from PyQt5.QtCore import (
    Qt,
    QVariant,
    QModelIndex,
    pyqtSignal,
    pyqtSlot,
    QPersistentModelIndex,
)
from PyQt5.QtWidgets import QFileDialog

# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


class Columns(AbstractColumns):
    _members = ["name", "nickname", "bitrate", "transmit"]


Columns.indexes = Columns.indexes()

bitrates = OrderedDict(
    [
        (1000000, "1 MBit/s"),
        (500000, "500 kBit/s"),
        (250000, "250 kBit/s"),
        (125000, "125 kBit/s"),
    ]
)

default_bitrate = 500000


def available_buses():
    valid = []

    for interface in can.interfaces.VALID_INTERFACES:
        if interface == "pcan":
            for n in range(1, 9):
                channel = "PCAN_USBBUS{}".format(n)
                try:
                    bus = can.interface.Bus(bustype=interface, channel=channel)
                except:
                    pass
                else:
                    device_number = bus.get_device_number()
                    bus.shutdown()
                    valid.append(
                        {
                            "interface": interface,
                            "channel": channel,
                            "changeable_bitrate": True,
                            "device_number": device_number,
                        }
                    )

            for n in range(1, 17):
                channel = "PCAN_LANBUS{}".format(n)
                try:
                    bus = can.interface.Bus(bustype=interface, channel=channel)
                except:
                    pass
                else:
                    bus.shutdown()
                    valid.append(
                        {
                            "interface": interface,
                            "channel": channel,
                            "changeable_bitrate": False,
                            "device_number": None,
                        }
                    )

        elif interface == "kvaser":
            # TODO: get the actual number of available devices rather
            #       than hard coding?
            #
            #       can.interfaces.kvaser.canGetNumberOfChannels())
            for channel in range(0, 8):
                try:
                    bus = can.interface.Bus(bustype=interface, channel=channel)
                except:
                    pass
                else:
                    bus.shutdown()
                    valid.append(
                        {
                            "interface": interface,
                            "channel": channel,
                            "changeable_bitrate": True,
                            "device_number": None,
                        }
                    )
        elif interface == "socketcan":
            for n in range(9):
                channel = "can{}".format(n)
                try:
                    bus = can.interface.Bus(bustype=interface, channel=channel)
                except:
                    pass
                else:
                    bus.shutdown()
                    valid.append(
                        {
                            "interface": interface,
                            "channel": channel,
                            "changeable_bitrate": False,
                            "device_number": None,
                        }
                    )
            for n in range(9):
                channel = "vcan{}".format(n)
                try:
                    bus = can.interface.Bus(bustype=interface, channel=channel)
                except:
                    pass
                else:
                    bus.shutdown()
                    valid.append(
                        {
                            "interface": interface,
                            "channel": channel,
                            "changeable_bitrate": False,
                            "device_number": None,
                        }
                    )
        else:
            print(
                "Availability check not implemented for {}".format(interface),
                file=sys.stderr,
            )

    return valid


class Bus(TreeNode):
    def __init__(
        self, interface, channel, changeable_bitrate=False, device_number=None
    ):
        TreeNode.__init__(self)

        self.interface = interface
        self.channel = channel
        self.changeable_bitrate = changeable_bitrate
        self.device_number = device_number

        self.bitrate = default_bitrate
        self.separator = " - "

        if self.interface is not None:
            name = self._generate_fields_name()
        else:
            name = "Offline"

        self.fields = Columns(
            name=name, nickname="-", bitrate=bitrates[self.bitrate], transmit=""
        )

        self._checked = Columns.fill(Qt.Unchecked)

        self.bus = epyqlib.busproxy.BusProxy(
            transmit=self.checked(Columns.indexes.transmit)
        )

    def terminate(self):
        self.bus.terminate()
        logging.debug("{} terminated".format(object.__repr__(self)))

    def set_data(self, data):
        for key, value in bitrates.items():
            if data == value:
                self.bitrate = key
                self.fields.bitrate = data

                self.set_bus()

        raise ValueError(
            "{} not found in {}".format(data, ", ".join(bitrates.values()))
        )

    def enumeration_strings(self, include_values=False):
        return bitrates.values()

    def unique(self):
        return "{} - {}".format(self.interface, self.channel)

    def append_child(self, child):
        TreeNode.append_child(self, child)

    def checked(self, column):
        return self._checked[column]

    def set_checked(self, checked, column):
        if column in [Columns.indexes.name, Columns.indexes.transmit]:
            if self.interface is None:
                self._checked[column] = Qt.Unchecked

                return

            self._checked[column] = checked

            if self._checked[column] == Qt.Checked:
                for device in self.children:
                    if device.checked(column) != Qt.Unchecked:
                        device.set_checked(checked=Qt.Checked, column=column)
            elif self._checked[column] == Qt.Unchecked:
                for device in self.children:
                    if device.checked(column) != Qt.Unchecked:
                        device.set_checked(checked=Qt.PartiallyChecked, column=column)

            if column == Columns.indexes.name:
                self.set_bus()
            elif column == Columns.indexes.transmit:
                self.bus.transmit = checked == Qt.Checked

    def set_bus(self):
        if self.interface is None:
            return

        self.bus.set_bus(None)

        if self._checked.name == Qt.Checked:
            real_bus = can.interface.Bus(
                bustype=self.interface, channel=self.channel, bitrate=self.bitrate
            )
            # TODO: Yuck, but it helps recover after connecting to a bus with
            #       the wrong speed.  So, find a better way.
            time.sleep(0.5)
        else:
            real_bus = None

        self.bus.set_bus(bus=real_bus)

    def set_nickname(self, name):
        self.fields.nickname = name

    def set_device_number(self, device_number):
        self.device_number = device_number
        self.fields.name = self._generate_fields_name()

    def _generate_fields_name(self):
        fields = [self.interface, self.channel, self.device_number]
        return self.separator.join(str(field) for field in fields if field is not None)


class Device(TreeNode):
    def __init__(self, device):
        TreeNode.__init__(self)

        self.fields = Columns(name=device.name, nickname="-", bitrate="", transmit="")

        self._checked = Columns.fill(Qt.Unchecked)

        self.device = device
        self.device.bus.transmit = self._checked.transmit == Qt.Checked
        self.name = ""

    def terminate(self):
        self.device.terminate()
        logging.debug("{} terminated".format(object.__repr__(self)))

    def unique(self):
        return self.device

    def checked(self, column):
        return self._checked[column]

    def set_checked(self, checked, column):
        if self.device.bus is None:
            return

        if column in [Columns.indexes.name, Columns.indexes.transmit]:
            if checked == Qt.Checked:
                if self.tree_parent.checked(column) == Qt.Checked:
                    self._checked[column] = Qt.Checked
                else:
                    if self._checked[column] == Qt.Unchecked:
                        self._checked[column] = Qt.PartiallyChecked
                    else:
                        self._checked[column] = Qt.Unchecked
            elif checked == Qt.PartiallyChecked:
                self._checked[column] = Qt.PartiallyChecked
            else:
                self._checked[column] = Qt.Unchecked

            self.device.bus_status_changed(
                online=self._checked.name == Qt.Checked,
                transmit=self._checked.transmit == Qt.Checked,
            )

            if column == Columns.indexes.name:
                if self._checked[column] == Qt.Unchecked:
                    self.device.bus.set_bus()
                else:
                    self.device.bus.set_bus(self.tree_parent.bus)

            elif column == Columns.indexes.transmit:
                self.device.bus.transmit = self._checked[column] == Qt.Checked

    def set_nickname(self, name):
        self.fields.nickname = name
        self.device.nickname = name


class Tree(TreeNode):
    def __init__(self):
        TreeNode.__init__(self)


class Model(epyqlib.pyqabstractitemmodel.PyQAbstractItemModel):
    device_removed = pyqtSignal(epyqlib.device.Device)
    details_changed = pyqtSignal()

    def __init__(self, root, parent=None):
        buses = [
            {
                "interface": None,
                "channel": None,
                "changeable_bitrate": False,
                "device_number": None,
            }
        ] + available_buses()
        for bus in buses:

            bus = Bus(
                interface=bus["interface"],
                channel=bus["channel"],
                changeable_bitrate=bus["changeable_bitrate"],
                device_number=bus["device_number"],
            )
            root.append_child(bus)
            went_offline = functools.partial(self.went_offline, node=bus)
            bus.bus.went_offline.connect(went_offline)

        editable_columns = Columns.fill(False)
        editable_columns.nickname = True
        editable_columns.bitrate = True

        checkbox_columns = Columns.fill(False)
        checkbox_columns.name = True
        checkbox_columns.transmit = True

        epyqlib.pyqabstractitemmodel.PyQAbstractItemModel.__init__(
            self,
            root=root,
            editable_columns=editable_columns,
            checkbox_columns=checkbox_columns,
            parent=parent,
        )

        self.headers = Columns(
            name="Bus/Device", nickname="Name", bitrate="Bitrate", transmit="Transmit"
        )

    def terminate(self):
        def terminate_devices(node, _):
            if isinstance(node, Device):
                node.terminate()

        def terminate_buses(node, _):
            if isinstance(node, Bus):
                node.terminate()

        self.root.traverse(terminate_devices, internal_nodes=True)
        self.root.traverse(terminate_buses, internal_nodes=True)
        logging.debug("{} terminated".format(object.__repr__(self)))

    def went_offline(self, node):
        # TODO: trigger gui update, or find a way that does it automatically
        node.set_checked(checked=Qt.Unchecked, column=Columns.indexes.name)
        self.changed(
            node, Columns.indexes.name, node, Columns.indexes.name, [Qt.CheckStateRole]
        )

    def setData(self, index, data, role=None):
        if index.column() == Columns.indexes.bitrate:
            if role == Qt.EditRole:
                node = self.node_from_index(index)
                try:
                    node.set_data(data)
                except ValueError:
                    return False
                self.dataChanged.emit(index, index)
                return True
        elif index.column() == Columns.indexes.nickname:
            if role == Qt.EditRole:
                node = self.node_from_index(index)
                try:
                    node.set_nickname(data)
                except ValueError:
                    return False
                self.dataChanged.emit(index, index)
                self.details_changed.emit()
                return True
        elif index.column() in [Columns.indexes.name, Columns.indexes.transmit]:
            if role == Qt.CheckStateRole:
                node = self.node_from_index(index)

                node.set_checked(checked=data, column=index.column())

                # TODO: CAMPid 9349911217316754793971391349
                children = len(node.children)
                if children > 0:
                    self.changed(
                        node.children[0],
                        Columns.indexes[0],
                        node.children[-1],
                        Columns.indexes[-1],
                        [Qt.CheckStateRole],
                    )

                return True

        return False

    def add_device(self, bus, device):
        index = len(bus.children)

        # TODO: move to TreeNode?
        self.begin_insert_rows(bus, index, index)
        bus.append_child(device)
        self.end_insert_rows()

        persistent_index = QPersistentModelIndex(self.index_from_node(bus))
        self.layoutChanged.emit([persistent_index])

    def remove_device(self, device):
        bus = device.tree_parent
        row = bus.children.index(device)

        device.device.bus.set_bus()

        self.begin_remove_rows(bus, row, row)
        bus.remove_child(row)
        self.end_remove_rows()

        persistent_index = QPersistentModelIndex(self.index_from_node(bus))
        self.layoutChanged.emit([persistent_index])

        self.device_removed.emit(device.device)

    def device_from_widget(self, widget):
        def check(node, matches, widget=widget):
            if isinstance(node, Device) and node.device.ui is widget:
                matches.append(node.device)

        matches = []
        self.root.traverse(call_this=check, payload=matches)

        if len(matches) == 0:
            return None

        (device,) = matches

        return device

    def flags(self, index):
        flags = super().flags(index)

        if index.column() == Columns.indexes.bitrate:
            node = self.node_from_index(index)
            if isinstance(node, Bus):
                if not node.changeable_bitrate:
                    flags &= ~Qt.ItemIsEnabled

        return flags


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



import contextlib
import glob
import os
import shutil
import zipfile

import pytest

with contextlib.suppress(ImportError):
    import epyqlib.collectdevices
import epyqlib.tests.common


@pytest.fixture
def zipped_customer_device_path(tmpdir):
    name = "customer"
    path = epyqlib.tests.common.devices["customer"]

    epyqlib.collectdevices.collect(
        devices={
            name: {"file": os.path.basename(path), "groups": ("a",)},
        },
        output_directory=tmpdir,
        dry_run=False,
        groups=("a",),
        device_path=path,
        in_repo=False,
    )

    return os.path.join(tmpdir, "{}.epz".format(name))


@pytest.fixture
def customer_device_path(tmpdir, zipped_customer_device_path):
    zip = zipfile.ZipFile(zipped_customer_device_path)
    zip.extractall(tmpdir)
    (path,) = glob.glob(os.path.join(tmpdir, "**", "*.epc"), recursive=True)

    return path



from PyQt5 import QtDesigner
import epyqlib.variableselectionview
import epyqlib.abstractpluginclass

# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


class VariableSelectionViewPlugin(epyqlib.abstractpluginclass.AbstractPlugin):
    def __init__(self, parent=None):
        epyqlib.abstractpluginclass.AbstractPlugin.__init__(self, parent=parent)

        self._group = "EPC - General"
        self._init = epyqlib.variableselectionview.VariableSelectionView
        self._module_path = "epyqlib.variableselectionview"
        self._name = "VariableSelectionView"



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

import epyqlib.widgets.abstracttxwidget
import epyqlib.widgets.enum_ui


# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


class Enum(epyqlib.widgets.abstracttxwidget.AbstractTxWidget):
    def __init__(self, parent=None, in_designer=False):
        self._frame = None
        self._signal = None

        super().__init__(
            ui_class=epyqlib.widgets.enum_ui.Ui_Form,
            parent=parent,
            in_designer=in_designer,
        )

        # TODO: CAMPid 398956661298765098124690765
        self.ui.value.currentTextChanged.connect(self.widget_value_changed)

    def set_value(self, value):
        if self.signal_object is not None:
            if len(self.signal_object.enumeration) > 0:
                value = self.signal_object.full_string
            else:
                value = self.signal_object.format_float()
        elif value is None:
            value = "-"
        else:
            # TODO: quit hardcoding this and it's better implemented elsewhere
            value = "{0:.2f}".format(value)

        self.ui.value.setCurrentText(value)

    def set_signal(self, signal=None, force_update=False):
        if signal is not self.signal_object or force_update:
            if self.ui is not None:
                self.ui.value.clear()
            if signal is not None:

                full_strings = []
                # TODO: CAMPid 94562754956589992752348667
                for value in sorted(signal.enumeration.keys()):
                    # TODO: CAMPid 85478672616219005471279
                    enum_string = signal.enumeration[value]
                    full_strings.append(
                        signal.enumeration_format_re["format"].format(
                            s=enum_string, v=value
                        )
                    )

                self.ui.value.addItems(full_strings)

        super().set_signal(signal, force_update=force_update)


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



import hashlib
import os
from os import path

from epyqlib.tabs.files.files_utils import ensure_dir


class FilesManager:
    def __init__(self, files_dir):
        cache_dir = path.join(files_dir, "files")
        ensure_dir(cache_dir)
        self._cache_dir = cache_dir

    def verify_cache(self):
        for filename in self.hashes():
            hash = self._md5(filename)
            if hash != filename:
                print(f"File {filename} failed hash verification. Deleting.")
                os.unlink(path.join(self._cache_dir, filename))

    def _md5(self, filename: str) -> str:
        md5 = hashlib.md5()
        with open(path.join(self._cache_dir, filename), "rb") as file:
            for chunk in iter(lambda: file.read(4096), b""):
                md5.update(chunk)
        return md5.hexdigest()

    def hashes(self):
        return os.listdir(self._cache_dir)

    def has_hash(self, hash: str) -> bool:
        return path.exists(path.join(self._cache_dir, hash))

    def get_file_path(self, filename: str):
        return path.join(self._cache_dir, filename)

    def stat(self, filename) -> os.stat_result:
        return os.stat(path.join(self._cache_dir, filename))

    def move_into_cache(self, file_path: str):
        basename = os.path.basename(file_path)
        return os.rename(file_path, path.join(self._cache_dir, basename))

    def delete_from_cache(self, hash: str):
        os.unlink(path.join(self._cache_dir, hash))



import pathlib

from PyQt5 import QtWidgets
from PyQt5.QtCore import QSortFilterProxyModel, Qt
import twisted.internet.threads

import epyqlib.cmemoryparser
import epyqlib.datalogger
import epyqlib.utils.qt
import epyqlib.utils.twisted
import epyqlib.variableselection_ui


# See file COPYING in this source tree
__copyright__ = "Copyright 2017, EPC Power Corp."
__license__ = "GPLv2+"


class VariableSelection(QtWidgets.QWidget):
    def __init__(self, parent=None, in_designer=False):
        super().__init__(parent=parent)

        self.in_designer = in_designer

        self.ui = epyqlib.variableselection_ui.Ui_Form()
        self.ui.setupUi(self)

        self.ui.load_binary_button.clicked.connect(self.load_binary)
        self.ui.save_selection_button.clicked.connect(self.save_selection)
        self.ui.load_selection_button.clicked.connect(self.load_selection)
        self.ui.update_parameters_button.clicked.connect(self.update_parameters)
        self.ui.process_raw_log_button.clicked.connect(self.process_raw_log)
        self.ui.process_raw_log_button.setEnabled(False)

        self.ui.view.ui.tree_view.setContextMenuPolicy(Qt.CustomContextMenu)
        self.ui.view.ui.tree_view.customContextMenuRequested.connect(self.context_menu)

        self.progress = None

    def set_signal_paths(
        self,
        reset_signal_path,
        recording_signal_path,
        configuration_is_valid_signal_path,
    ):
        self.ui.reset_button.set_signal_path(reset_signal_path)
        self.ui.logging_led.set_signal_path(recording_signal_path)
        self.ui.configuration_is_valid_led.set_signal_path(
            configuration_is_valid_signal_path
        )

    def set_model(self, model):
        self.ui.view.set_model(model)

    def set_sorting_enabled(self, enabled):
        self.ui.view.set_sorting_enabled(enabled)

    def sort_by_column(self, column, order):
        self.ui.view.sort_by_column(column=column, order=order)

    def save_selection(self):
        filters = [("EPC Variable Selection", ["epv"]), ("All Files", ["*"])]
        filename = epyqlib.utils.qt.file_dialog(filters, save=True, parent=self)

        if filename is not None:
            model = self.nonproxy_model()
            model.save_selection(filename=filename)

    def load_selection(self):
        filters = [("EPC Variable Selection", ["epv"]), ("All Files", ["*"])]
        filename = epyqlib.utils.qt.file_dialog(filters, parent=self)

        if filename is not None:
            model = self.nonproxy_model()
            model.load_selection(filename=filename)

    # TODO: CAMPid 07943342700734207878034207087
    def nonproxy_model(self):
        model = self.ui.view.model
        while isinstance(model, QSortFilterProxyModel):
            model = model.sourceModel()

        return model

    def load_binary(self):
        filters = [("TICOFF Binaries", ["out"]), ("All Files", ["*"])]
        filename = epyqlib.utils.qt.file_dialog(filters, parent=self)

        if filename is not None:
            # TODO: CAMPid 9632763567954321696542754261546
            self.progress = epyqlib.utils.qt.progress_dialog(parent=self)
            self.progress.setLabelText("Loading binary...")

            model = self.nonproxy_model()

            self.progress.show()

            d = twisted.internet.threads.deferToThread(
                epyqlib.cmemoryparser.process_file, filename=filename
            )
            d.addCallback(model.update_from_loaded_binary)
            d.addCallback(
                epyqlib.utils.twisted.detour_result,
                self.ui.process_raw_log_button.setEnabled,
                True,
            )
            d.addBoth(epyqlib.utils.twisted.detour_result, self.progress_cleanup)
            d.addErrback(epyqlib.utils.twisted.errbackhook)

    def progress_cleanup(self):
        self.progress.close()
        self.progress.deleteLater()
        self.progress = None

    def update_parameters(self):
        model = self.nonproxy_model()
        model.update_parameters(parent=self)

    def process_raw_log(self):
        filters = [("Raw", ["raw"]), ("All Files", ["*"])]
        raw_filename = epyqlib.utils.qt.file_dialog(filters, parent=self)

        if raw_filename is not None:
            filters = [("CSV", ["csv"]), ("All Files", ["*"])]
            csv_guess = str(
                pathlib.Path(raw_filename).with_suffix("." + filters[0][1][0])
            )
            csv_filename = epyqlib.utils.qt.file_dialog(
                filters, save=True, parent=self, dir=csv_guess
            )

            if csv_filename is not None:
                with open(raw_filename, "rb") as f:
                    data = f.read()

                model = self.nonproxy_model()

                self.progress = epyqlib.utils.qt.progress_dialog(parent=self)
                self.progress.setLabelText("Processing Raw Log...")

                self.progress.show()

                d = model.parse_log(data=data, csv_path=csv_filename)
                d.addBoth(epyqlib.utils.twisted.detour_result, self.progress_cleanup)
                d.addErrback(epyqlib.utils.twisted.errbackhook)

    def context_menu(self, position):
        index = self.ui.view.ui.tree_view.indexAt(position)
        index = self.ui.view.ui.tree_view.model().mapToSource(index)

        if not index.isValid():
            return

        node = self.nonproxy_model().node_from_index(index)

        menu = QtWidgets.QMenu()
        read_action = menu.addAction("Read")

        action = menu.exec(self.ui.view.ui.tree_view.viewport().mapToGlobal(position))

        if action is None:
            pass
        elif action is read_action:
            self.nonproxy_model().read(variable=node)



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

import enum
import struct
import socket

from ctypes import *


# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


# TODO: traceability?
WEHMD_API_SOCKET = "/var/run/wehmd_api"


class Wehmd:
    def __init__(self):
        self.socket = socket.socket(family=socket.AF_UNIX, type=socket.SOCK_STREAM)
        self.socket.connect(WEHMD_API_SOCKET)

    def read_boot_mode(self):
        message = wehmd_msg_t()
        message.command = int(CommandType.WEHMD_CMD_GET_EE_BOOT_MODE)
        message.data.bootMode = 0

        self.send(message)
        response = self.receive()
        return wehmd_msg_t.from_buffer_copy(response)

    def write_boot_mode(self, value):
        message = wehmd_msg_t()
        message.command = int(CommandType.WEHMD_CMD_SET_EE_BOOT_MODE)
        message.data.bootMode = value

        self.send(message)
        response = self.receive()
        return wehmd_msg_t.from_buffer_copy(response)

    def send(self, message):
        raw = (c_char * sizeof(message))()
        memmove(raw, byref(message), sizeof(message))

        total_sent = 0
        while total_sent < len(raw):
            sent = self.socket.send(raw[total_sent:])
            if sent == 0:
                raise RuntimeError("socket connection broken")
            total_sent = total_sent + sent

    def receive(self):
        chunks = []
        bytes_received = 0
        while bytes_received < sizeof(wehmd_msg_t):
            chunk = self.socket.recv(min(sizeof(wehmd_msg_t) - bytes_received, 2048))
            if chunk == b"":
                raise RuntimeError("socket connection broken")
            chunks.append(chunk)
            bytes_received = bytes_received + len(chunk)
        return b"".join(chunks)


CommandType = enum.IntEnum(
    "CommandType",
    """
    WEHMD_CMD_GET_OTC
    WEHMD_CMD_OTC
    WEHMD_CMD_GET_EE_PRODUCT_TYPE
    WEHMD_CMD_EE_PRODUCT_TYPE
    WEHMD_CMD_GET_EE_PRODUCT_VARIANT
    WEHMD_CMD_EE_PRODUCT_VARIANT
    WEHMD_CMD_GET_EE_BOARD_REVISION
    WEHMD_CMD_EE_BOARD_REVISION
    WEHMD_CMD_GET_EE_BOARD_NAME
    WEHMD_CMD_EE_BOARD_NAME
    WEHMD_CMD_GET_EE_PART_NUMBER
    WEHMD_CMD_EE_PART_NUMBER
    WEHMD_CMD_GET_EE_SERIAL_NUMBER
    WEHMD_CMD_EE_SERIAL_NUMBER
    WEHMD_CMD_GET_EE_BOOT_MODE
    WEHMD_CMD_EE_BOOT_MODE
    WEHMD_CMD_GET_EE_UNIQUE_RELEASE_ID
    WEHMD_CMD_EE_UNIQUE_RELEASE_ID
    WEHMD_CMD_GET_EE_DELETE_DIAG
    WEHMD_CMD_EE_DELETE_DIAG
    WEHMD_CMD_GET_EE_CONSOLE_MODE
    WEHMD_CMD_EE_CONSOLE_MODE
    WEHMD_CMD_GET_API_VERSION
    WEHMD_CMD_API_VERSION
    WEHMD_CMD_ERROR
    WEHMD_CMD_SET_EE_CONSOLE_MODE
    WEHMD_CMD_GET_EE_DISPLAY_BRIGHTNESS
    WEHMD_CMD_SET_EE_DISPLAY_BRIGHTNESS
    WEHMD_CMD_EE_DISPLAY_BRIGHTNESS
    WEHMD_CMD_GET_EE_KEYPAD_BRIGHTNESS
    WEHMD_CMD_SET_EE_KEYPAD_BRIGHTNESS
    WEHMD_CMD_EE_KEYPAD_BRIGHTNESS
    WEHMD_CMD_SET_EE_BOOT_MODE
    WEHMD_CMD_SET_EE_RAM_TEST
    WEHMD_CMD_GET_EE_RAM_TEST
    WEHMD_CMD_EE_RAM_TEST
    WEHMD_CMD_GET_EE_KEYPAD_COLOR
    WEHMD_CMD_SET_EE_KEYPAD_COLOR
    WEHMD_CMD_EE_KEYPAD_COLOR
""",
    start=0,
)


wehmd_cmd_t = c_uint
wehmd_err_type_t = c_uint


class wehmd_error_t(Structure):
    _fields_ = [("command", wehmd_cmd_t), ("errorType", wehmd_err_type_t)]


class wehmd_msg_data_t(Union):
    _fields_ = [
        ("otc", c_uint),
        ("productType", c_ubyte),
        ("productVariant", c_ubyte),
        ("boardRevision", c_ubyte),
        ("boardName", c_char * 31),
        ("partNumber", c_char * 21),
        ("serialNumber", c_char * 21),
        ("bootMode", c_ubyte),
        ("uniqueReleaseID", c_uint),
        ("deleteDiag", c_ubyte),
        ("consoleMode", c_ubyte),
        ("apiVersion", c_uint),
        ("displayBrightness", c_ubyte),
        ("keypadBrightness", c_ubyte),
        ("keypadColor", c_uint),
        ("ramTest", c_ubyte),
        ("error", wehmd_error_t),
    ]


class wehmd_msg_t(Structure):
    _fields_ = [("command", wehmd_cmd_t), ("data", wehmd_msg_data_t)]


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

from PyQt5.QtCore import pyqtProperty, Qt
from PyQt5.QtGui import QFocusEvent, QKeyEvent

import epyqlib.utils.qt
import epyqlib.widgets.abstracttxwidget
import epyqlib.widgets.epc_ui


# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


class Epc(epyqlib.widgets.abstracttxwidget.AbstractTxWidget):
    def __init__(
        self,
        parent=None,
        ui_class=epyqlib.widgets.epc_ui.Ui_Form,
        in_designer=False,
    ):
        self._frame = None
        self._signal = None

        self._show_enumeration_value = True

        self.isBlue = False  # By default, the text color is black.

        super().__init__(
            ui_class=ui_class,
            parent=parent,
            in_designer=in_designer,
        )

        # TODO: CAMPid 398956661298765098124690765
        self.ui.value.returnPressed.connect(self.widget_value_changed)
        self.ui.value.installEventFilter(self)

        self.ui.edit_button.hide()

        self.update()

    @pyqtProperty(bool)
    def show_enumeration_value(self):
        return self._show_enumeration_value

    @show_enumeration_value.setter
    def show_enumeration_value(self, show):
        self._show_enumeration_value = bool(show)

        self.update()

    def set_signal(self, *args, **kwargs):
        super().set_signal(*args, **kwargs)

        self.update()

    def cancel_edit(self):
        self.signal_object.set_human_value(
            self.signal_object.get_human_value(), force=True
        )

    def eventFilter(self, object, event):
        super_result = super().eventFilter(object, event)

        if not super_result:
            if object is self.ui.value:
                if isinstance(event, QKeyEvent):
                    if event.key() == Qt.Key_Escape:
                        # TODO: for some reason this is triggering 3x
                        #       for one press
                        self.cancel_edit()

                if isinstance(event, QFocusEvent):
                    if event.lostFocus():
                        self.cancel_edit()

        return super_result

    # TODO: CAMPid 097327143264214321432453216453762354
    def update(self):
        width = self.calculate_max_value_width()
        if width is not None:
            value = self.ui.value
            value.setMinimumWidth(width)

        if self.signal_object is not None:
            alignment = (
                Qt.AlignRight
                if len(self.signal_object.enumeration) == 0
                else Qt.AlignCenter
            )
            alignment |= Qt.AlignVCenter
            self.ui.value.setAlignment(alignment)

    # TODO: CAMPid 989849193479134917954791341
    def calculate_max_value_width(self):
        if self.signal_object is None:
            return None

        signal = self.signal_object

        longer = max(
            [signal.format_float(v) for v in [signal.min, signal.max]], key=len
        )

        digits = len(longer)

        if "." in longer:
            decimal = "."
            digits -= 1
        else:
            decimal = ""

        self.ui.value.setVisible(self.ui.value.isVisibleTo(self))
        metric = self.ui.value.fontMetrics()
        chars = ["{:}".format(i) for i in range(10)]
        widths = [metric.width(c) for c in chars]
        widest_width = max(widths)
        widest_char = chars[widths.index(widest_width)]
        string = "{}".format((widest_char * digits) + decimal)

        strings = signal.enumeration_strings()
        strings.append(string)

        # TODO: really figure out the spacing needed but for now
        #       just add a space on each side for buffer space
        strings = [s + "   " for s in strings]

        return max([metric.width(s) for s in strings])

    def widget_value_changed(self):
        super().widget_value_changed(self.ui.value.text())

    def set_value(self, value):
        if self.signal_object is not None:
            if len(self.signal_object.enumeration) > 0:
                value = (
                    self.signal_object.full_string
                    if self.show_enumeration_value
                    else self.signal_object.enumeration_text
                )
            else:
                value = self.signal_object.short_string
        elif value is None:
            # TODO: quit hardcoding this and it's better implemented elsewhere
            value = "{0:.2f}".format(0)
        else:
            # TODO: quit hardcoding this and it's better implemented elsewhere
            value = "{0:.2f}".format(value)

        self.ui.value.setText(value)

    @epyqlib.widgets.abstracttxwidget.AbstractTxWidget.tx.setter
    def tx(self, tx):
        epyqlib.widgets.abstracttxwidget.AbstractTxWidget.tx.fset(self, tx)

        self.ui.value.acceptsDrops = self.tx

        self.ui.value.setMouseTracking(self.tx)
        self.ui.value.setReadOnly(not self.tx)
        self.ui.value.setFocusPolicy(Qt.StrongFocus if self.tx else Qt.NoFocus)

        self.update()

    # Determines if the text for the label should be black or blue.
    @pyqtProperty(bool)
    def textColorBlue(self):
        return self.isBlue

    @textColorBlue.setter
    def textColorBlue(self, value):
        self.isBlue = value

        if self.isBlue:
            self.ui.label.setStyleSheet("QLabel { color : blue; }")
        else:
            self.ui.label.setStyleSheet("")


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

from PyQt5.QtCore import pyqtProperty
from PyQt5 import QtWidgets

import epyqlib.iopoint_ui
import epyqlib.utils.qt


# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


class IoPoint(QtWidgets.QWidget):
    def __init__(self, parent=None, in_designer=False):
        super().__init__(parent=parent)

        self.in_designer = in_designer

        self.ui = epyqlib.iopoint_ui.Ui_Form()
        self.ui.setupUi(self)

        self.ui.status.in_designer = self.in_designer
        self.ui.set.in_designer = self.in_designer
        self.ui.override.in_designer = self.in_designer

        self.update_configuration()

    @pyqtProperty(bool)
    def tx(self):
        return self.ui.set.tx

    @tx.setter
    def tx(self, tx):
        self.ui.set.tx = bool(tx)
        self.ui.override.tx = bool(tx)

        self.update_configuration()

    @pyqtProperty(str)
    def label_override(self):
        return self.ui.status.label_override

    @label_override.setter
    def label_override(self, label):
        self.ui.status.label_override = label

        self.update_configuration()

        # TODO: if not empty then do something

    @pyqtProperty("QString")
    def status_signal_path(self):
        return self.ui.status.signal_path

    @status_signal_path.setter
    def status_signal_path(self, value):
        self.ui.status.signal_path = value
        self.update_configuration()

    @pyqtProperty("QString")
    def set_signal_path(self):
        return self.ui.set.signal_path

    @set_signal_path.setter
    def set_signal_path(self, value):
        self.ui.set.signal_path = value
        self.update_configuration()

    @pyqtProperty("QString")
    def override_signal_path(self):
        return self.ui.override.signal_path

    @override_signal_path.setter
    def override_signal_path(self, value):
        self.ui.override.signal_path = value
        self.update_configuration()

    def update_configuration(self):
        self.ui.set.setVisible(self.tx)
        self.ui.override.setVisible(self.tx)
        if self.tx:
            self.ui.set_label.show()
            self.ui.override_label.show()
        else:
            self.ui.set_label.hide()
            self.ui.override_label.hide()


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

import epyqlib.widgets.abstractwidget

from PyQt5.QtCore import pyqtProperty, pyqtSlot, QRectF, Qt
from PyQt5.QtGui import QPainter, QPen, QColor

# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


def arc_angle(angle):
    # Qt expects 1/16 of a degree
    # http://doc.qt.io/qt-5/qpainter.html#drawArc

    return round(16 * angle)


class RingBar(epyqlib.widgets.abstractwidget.AbstractWidget):
    def __init__(self, parent=None, in_designer=False):
        self.in_designer = in_designer
        epyqlib.widgets.abstractwidget.AbstractWidget.__init__(self, parent=parent)

        self._background_color = QColor("#474747")
        self._clockwise = False
        self._color = QColor("#39C550")
        self._maximum = 100
        self._minimum = 0
        self._thickness = 10
        self._value = 75

        self._zero_angle = 0
        self._angle_span = 360

        self._override_range = False

    # TODO  use actual OverrideRange mixin
    @pyqtProperty(bool)
    def override_range(self):
        return self._override_range

    @override_range.setter
    def override_range(self, override):
        self._override_range = bool(override)

    @pyqtProperty(bool)
    def clockwise(self):
        return self._clockwise

    @clockwise.setter
    def clockwise(self, clockwise):
        self._clockwise = clockwise

        self.update()

    @pyqtProperty(float)
    def zero_angle(self):
        return self._zero_angle

    @zero_angle.setter
    def zero_angle(self, angle):
        self._zero_angle = angle

        self.update()

    @pyqtProperty(float)
    def angle_span(self):
        return self._angle_span

    @angle_span.setter
    def angle_span(self, angle):
        self._angle_span = angle

        self.update()

    @pyqtProperty(float)
    def thickness(self):
        return self._thickness

    @thickness.setter
    def thickness(self, thickness):
        self._thickness = thickness

        self.update()

    @pyqtProperty("QColor")
    def color(self):
        return self._color

    @color.setter
    def color(self, color):
        self._color = QColor(color)

        self.update()

    @pyqtProperty("QColor")
    def background_color(self):
        return self._background_color

    @background_color.setter
    def background_color(self, color):
        self._background_color = QColor(color)

        self.update()

    def dimension(self):
        return min(self.width(), self.height())

    def resizeEvent(self, event):
        epyqlib.widgets.abstractwidget.AbstractWidget.resizeEvent(self, event)

        self.update_layout()

    def tweaked_thickness(self):
        return self.thickness + 1

    def paintEvent(self, event):
        epyqlib.widgets.abstractwidget.AbstractWidget.paintEvent(self, event)

        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing, True)
        painter.translate(self.width() / 2, self.height() / 2)

        diameter = self.dimension() - self.tweaked_thickness()
        radius = diameter / 2

        rectangle = QRectF(-radius, -radius, diameter, diameter)

        maximum = max(abs(self.maximum), abs(self.minimum))

        span_angle = self.angle_span * (self.value / maximum)

        if self.clockwise:
            span_angle = -span_angle

        pen = QPen()
        pen.setWidthF(self.tweaked_thickness())
        pen.setCapStyle(Qt.RoundCap)

        pen.setColor(self.background_color)
        painter.setPen(pen)
        painter.drawEllipse(rectangle)

        pen.setColor(self.color)
        painter.setPen(pen)

        qt_span = arc_angle(span_angle)
        if qt_span == 0:
            qt_span = 1

        painter.drawArc(rectangle, arc_angle(self.zero_angle), qt_span)

    def update_layout(self):
        horizontal_margin = (self.width() - self.dimension()) / 2
        vertical_margin = (self.height() - self.dimension()) / 2

        horizontal_margin += self.thickness
        vertical_margin += self.thickness

        self.setContentsMargins(
            horizontal_margin, vertical_margin, horizontal_margin, vertical_margin
        )

    def update(self):
        self.update_layout()

        epyqlib.widgets.abstractwidget.AbstractWidget.update(self)

    @pyqtSlot(float)
    def set_value(self, value):
        self.value = value

    @pyqtProperty(float)
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value

        self.update()

    @pyqtProperty(float)
    def maximum(self):
        return self._maximum

    @maximum.setter
    def maximum(self, maximum):
        self._maximum = maximum

        self.update()

    @pyqtProperty(float)
    def minimum(self):
        return self._minimum

    @minimum.setter
    def minimum(self, minimum):
        self._minimum = minimum

        self.update()

    # TODO: CAMPid 2397847962541678243196352195498
    def set_range(self, min=None, max=None):
        # TODO: stop using min/max
        minimum = min
        maximum = max

        if self.override_range:
            minimum = self.minimum
            maximum = self.maximum

        if minimum == maximum:
            # TODO: pick the right exception
            raise Exception("Min and max may not be the same")
        elif minimum > maximum:
            # TODO: pick the right exception
            raise Exception("Min must be less than max")

        if minimum is not None:
            self.minimum = minimum
        if maximum is not None:
            self.maximum = maximum

    def set_unit_text(self, units):
        pass


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



import json
import os

import appdirs
from epyqlib.tabs.files.files_utils import ensure_dir


class Vars:
    auto_sync = "auto_sync"
    offline_mode = "offline_mode"
    refresh_token = "refresh_token"


class SyncConfig:
    _instance = None
    _tag = f"[{__name__}]"

    default_values = {Vars.auto_sync: True, Vars.offline_mode: False}

    def __init__(self, directory=None, filename="epyq-config.json"):
        self.required_keys = [
            key for key in Vars.__dict__.keys() if not key.startswith("__")
        ]

        self.config_dir = directory or appdirs.user_config_dir("Epyq", "EPC Power")
        self.cache_dir = directory or os.path.join(os.getcwd(), "sync")

        print(f"{self._tag} Using config dir: {self.config_dir}")
        print(f"{self._tag} Using cache dir: {self.cache_dir}")

        for dir in (self.config_dir, self.cache_dir):
            ensure_dir(dir)

        self.filename = os.path.join(self.config_dir, filename)
        self.file_error = False
        if os.path.exists(self.filename):
            self._read_file()
        else:
            self.config = dict()
            for key in self.required_keys:
                self.config[key] = self.default_values.get(key)

        self._write_file()

    @staticmethod
    def get_env() -> str:
        return "client"
        # return "internal"

    @staticmethod
    def get_instance():
        if SyncConfig._instance is None:
            SyncConfig._instance = SyncConfig()

        return SyncConfig._instance

    def _read_file(self):
        with open(self.filename, "r") as infile:
            contents = json.load(infile)
        if type(contents) is not dict:
            self.file_error = True
            raise ConfigurationError(
                "Configuration file is not a valid configuration JSON object."
            )
        for key in self.required_keys:
            if key not in contents.keys() or contents[key] is None:
                contents[key] = self.default_values.get(key)
        self.config = contents

    def _write_file(self):
        if not self.file_error:
            with open(self.filename, "w") as outfile:
                json.dump(self.config, outfile, indent=2)

    def get(self, key):
        return self.config.get(key)

    def get_bool(self, key):
        return self.get(key) or False

    def set(self, key, value):
        self.config[key] = value
        self._write_file()


class ConfigurationError(Exception):
    pass



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

import functools
import textwrap
import time

from PyQt5 import QtWidgets

import epyqlib.parameteredit_ui
import epyqlib.utils.qt


# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


class ParameterEdit(QtWidgets.QWidget):
    def __init__(
        self,
        parent=None,
        in_designer=False,
        edit=None,
        nv=None,
        dialog=None,
        esc_action=None,
    ):
        super().__init__(parent=parent)

        self.in_designer = in_designer

        self._edit = edit
        self._dialog = dialog
        self.nv = nv

        self.ui = epyqlib.parameteredit_ui.Ui_Form()
        self.ui.setupUi(self)

        self.ui.to_device.set_signal(self.nv)
        self.nv.status_signal.value_changed.connect(self.nv.value_changed)

        self.ui.save_to_nv_button.clicked.connect(self.save_to_nv)

        self.ui.to_device.edited.connect(self.edited)

        self.ui.description.setText(self.ui.to_device.toolTip())

        self.esc_action = esc_action
        self.ui.esc_button.clicked.connect(self.esc)

    def esc(self, checked):
        if self.esc_action is not None:
            self.esc_action()

    def edited(self, value):
        self.nv.read_from_device()
        time.sleep(0.05)
        self.nv.set_human_value(value)
        self.nv.write_to_device()

    def save_to_nv(self):
        focus_self = functools.partial(self.parent().setCurrentWidget, self)

        # TODO: CAMPid 93849811216123127753953680713426
        def inverter_to_nv():
            self.nv.tree_parent.module_to_nv()
            focus_self()

        self._dialog.focus(
            ok_action=inverter_to_nv,
            cancel_action=focus_self,
            label=textwrap.dedent(
                """\
                             Save all parameters to NV?"""
            ),
        )


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



import argparse
import itertools
import json
import logging
import sys

import dulwich.repo


def build_relatives(repo):
    children = {}
    parents = {}

    heads = {n: s for n, s in repo.get_refs().items() if n.startswith(b"refs/heads/")}

    for entry in repo.get_walker(include=list(heads.values())):
        for p in entry.commit.parents:
            parents.setdefault(entry.commit.id, set()).add(p)
            children.setdefault(p, set()).add(entry.commit.id)

    return children, parents


# def traverse(self, call_this, payload=None):
#     child = None
#     for child in self.children:
#         child.traverse(call_this, payload)
#
#     if child is None:
#         call_this(self, payload)


def interface_rev(repo, commit):
    tree = repo[commit.tree]
    _, swrevs_sha = tree.lookup_path(lambda sha: repo[sha], b"interface/swRevs.h")
    swrevs = repo[swrevs_sha]

    for line in swrevs.data.decode().splitlines():
        if line.startswith("#define INTERFACE_SW_REV "):
            return int(line.split("#define INTERFACE_SW_REV ")[1])

    return None


def matches_rev(target_rev, repo, commit):
    return target_rev == interface_rev(repo, commit)


def history(repo, base_commit, mappings, test, visited=None):
    if visited is None:
        visited = set()

    revisions = {base_commit}

    relatives = (mapping[base_commit.id] for mapping in mappings)
    relatives = itertools.chain(*relatives)

    for commit in relatives:
        if commit in visited:
            continue
        visited.add(commit)

        commit = repo[commit]
        if test(commit):
            revisions |= history(
                repo=repo,
                base_commit=commit,
                mappings=mappings,
                test=test,
                visited=visited,
            )

    return revisions


def contiguous_commits(repo, base_commit, test):
    children, parents = build_relatives(repo=repo)

    return history(
        repo=repo, base_commit=base_commit, mappings=(parents, children), test=test
    )


def parse_args(args):
    parser = argparse.ArgumentParser()
    # parser.add_argument('--verbose', '-v', action='count', default=0)
    parser.add_argument("--sha", "-s", required=True)
    parser.add_argument("--repository", "--repo", "-r", default=".")
    parser.add_argument("--json", "-j", action="store_true")

    return parser.parse_args(args)


def main(*args, logger):
    args = parse_args(args=args)

    encoded_sha = args.sha.encode()

    repo = dulwich.repo.Repo(args.repository)

    relatives = itertools.chain(*build_relatives(repo=repo))
    found = {relative for relative in relatives if relative.startswith(encoded_sha)}

    if len(found) == 0:
        raise Exception("No matches found for sha {}".format(args.sha))
    elif len(found) > 1:
        raise Exception(
            "Multiple matches found for sha {}:\n\n{}".format(
                args.sha,
                "".join("\n\t" + s for s in (sha.decode() for sha in sorted(found))),
            )
        )

    [sha] = found

    base_commit = repo[sha]
    target_rev = interface_rev(repo, base_commit)
    cc = contiguous_commits(
        repo=repo,
        base_commit=base_commit,
        test=lambda commit: matches_rev(
            target_rev=target_rev, repo=repo, commit=commit
        ),
    )

    h = sorted(cc, key=lambda c: c.commit_time, reverse=True)
    h = (c.id.decode() for c in h)

    if args.json:
        print(json.dumps(list(h), indent=4))
    else:
        print("\n".join(h))


def _entry_point():
    import traceback

    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    stream_handler = logging.StreamHandler()
    file_handler = logging.FileHandler("contiguouscommits.log")

    for handler in (stream_handler, file_handler):
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    logging.basicConfig(format="%(asctime)s - %(levelname)s - %(message)s")

    return main(*sys.argv[1:], logger=logger)
